# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import Qt

from karabo.common.scenemodel.api import EditableComboBoxModel
from karabo.common.states import State
from karabo.native import Configurable, Int32, String
from karabogui.binding.api import build_binding
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..combobox import EditableComboBox


class Object(Configurable):
    state = String(defaultValue=State.ON)
    prop = String(options=["foo", "bar", "baz", "qux"],
                  allowedStates=[State.ON])


class Other(Configurable):
    prop = Int32(options=[1, 2, 3, 5, 8])


@pytest.fixture
def combobox_setup(gui_app):
    # setup
    proxy = get_class_property_proxy(Object.getClassSchema(), "prop")
    controller = EditableComboBox(proxy=proxy, model=EditableComboBoxModel())
    controller.create(None)
    assert controller.widget is not None
    yield controller, proxy
    # teardown
    controller.destroy()
    assert controller.widget is None


def test_combobox_allowed(combobox_setup):
    controller, proxy = combobox_setup
    set_proxy_value(proxy, "state", "CHANGING")
    assert controller.widget.isEnabled() is False
    set_proxy_value(proxy, "state", "ON")
    assert controller.widget.isEnabled() is True


def test_combobox_set_value(combobox_setup):
    controller, proxy = combobox_setup
    set_proxy_value(proxy, "prop", "bar")
    assert controller.widget.currentIndex() == 1

    set_proxy_value(proxy, "prop", "nobar")
    assert controller.widget.currentIndex() == -1

    set_proxy_value(proxy, "prop", "baz")
    assert controller.widget.currentIndex() == 2


def test_combobox_edit_value(combobox_setup):
    controller, proxy = combobox_setup
    controller.widget.setCurrentIndex(3)
    assert proxy.edit_value == "qux"


def test_combobox_schema_update(gui_app):
    proxy = get_class_property_proxy(Other.getClassSchema(), "prop")
    controller = EditableComboBox(proxy=proxy)
    controller.create(None)
    assert controller.widget.currentIndex() == -1
    set_proxy_value(proxy, "prop", 2)
    assert controller.widget.currentIndex() == 1

    assert controller.widget.count() == 5

    build_binding(Object.getClassSchema(),
                  existing=proxy.root_proxy.binding)
    assert controller.widget.currentIndex() == -1
    assert controller.widget.count() == 4

    controller.destroy()


def test_combobox_focus(combobox_setup):
    controller, _ = combobox_setup
    combobox = controller.widget
    assert combobox.focusPolicy() == Qt.StrongFocus



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on February 16, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from enum import Enum

from qtpy.QtCore import Signal, Slot
from qtpy.QtWidgets import (
    QAction, QFrame, QHBoxLayout, QSizePolicy, QVBoxLayout, QWidget)

from karabogui import icons
from karabogui.dialogs.logbook_preview import LogBookPreview
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.util import generateObjectName
from karabogui.widgets.toolbar import ToolBar


class PanelActions(Enum):
    """Actions which can be done to all panels.
    """
    Maximize = 0
    Minimize = 1
    Dock = 2
    Undock = 3


class BasePanelWidget(QFrame):
    """The base class for all panels shown in the GUI. Panels can either be
    independent top-level windows or tabs within the main window of the Karabo
    gui.
    """
    signalPanelClosed = Signal(str)

    def __init__(self, title, allow_closing=False):
        super().__init__(parent=None)
        self.setWindowTitle(title)
        self.setFrameStyle(QFrame.Box | QFrame.Plain)
        self.setLineWidth(1)

        self.index = -1
        self.allow_closing = allow_closing
        self.panel_container = None
        self.is_docked = False
        self.tab_text_color = None

        self._fill_panel()

    # --------------------------------------
    # BasePanelWidget interface

    def get_content_widget(self):
        """Returns a QWidget containing the main content of the panel.
        """
        raise NotImplementedError

    def dock(self):
        """Called before this panel is docked into the main window.
        """

    def undock(self, maximized=False):
        """Called before this panel is undocked from the main window.
        """

    def maximize(self):
        """Called before this panel is maximized to fill the main window.
        """

    def minimize(self):
        """Called before this panel is minimized after filling the main window.
        """

    def update_tab_text_color(self, color=None):
        """Called when this panels tab text color should be changed.
        """
        self.tab_text_color = color
        self.panel_container.update_tab_text_color(self, self.tab_text_color)

    def toolbars(self):
        """This should create and return a list of `ToolBar` instances needed
        by this panel.
        """
        return []

    def info(self):
        """Retrieve the info for the KaraboLogBook"""

    # --------------------------------------
    # public methods

    def attach_to_container(self, container):
        if container is not None:
            self.is_docked = True
        else:
            self.is_docked = False

        self.panel_container = container
        # Set the toolbar visibility based on whether we're attached or not
        self.standard_toolbar.setVisible(container is not None)
        self.toolbar.setVisible(container is not None)

    def set_title(self, name):
        """Set the title of this panel whether its docked or undocked.
        """
        # Set the window title
        self.setWindowTitle(name)
        # And (maybe) the tab label
        if self.is_docked:
            panel = self.panel_container.widget(self.index)
            if panel is self:
                self.panel_container.setTabText(self.index, name)

    # --------------------------------------
    # Qt slots and callbacks

    @Slot()
    def handleLogBookPreview(self):
        dialog = LogBookPreview(parent=self)
        dialog.show()

    def closeEvent(self, event):
        if not self.allow_closing and self.panel_container is not None:
            self.onDock()
            event.ignore()
        else:
            if self.close():
                event.accept()
                if self.panel_container is not None and not self.is_docked:
                    # We are undocked and notify!
                    self.panel_container.removeUndockedPanel(self)
            else:
                event.ignore()

    def onUndock(self):
        self._update_toolbar_buttons(PanelActions.Undock)
        is_maximized = self.panel_container.maximized
        self.undock(is_maximized)
        self.panel_container.undock(self)

    def onDock(self):
        self._update_toolbar_buttons(PanelActions.Dock)
        self.dock()
        self.panel_container.dock(self)

    def onMaximize(self):
        self._update_toolbar_buttons(PanelActions.Maximize)
        self.maximize()
        i = self.panel_container.count()
        while i > -1:
            i -= 1
            w = self.panel_container.widget(i)
            if w != self:
                self.panel_container.removeTab(i)

        broadcast_event(KaraboEvent.MaximizePanel,
                        {'container': self.panel_container})
        self.panel_container.maximized = True

    def onMinimize(self):
        self._update_toolbar_buttons(PanelActions.Minimize)
        self.minimize()
        self.panel_container.insert_panels_after_maximize(self.index)

        broadcast_event(KaraboEvent.MinimizePanel,
                        {'container': self.panel_container})
        self.panel_container.maximized = False

    # --------------------------------------
    # private methods

    def _build_standard_toolbar(self):
        """This toolbar is shown by all panels which are attached to a
        container.
        """

        text = "Unpin as individual window"
        self.acUndock = QAction(icons.undock, "&Undock", self)
        self.acUndock.setToolTip(text)
        self.acUndock.setStatusTip(text)
        self.acUndock.triggered.connect(self.onUndock)

        text = "Pin this window to main program"
        self.acDock = QAction(icons.dock, "&Dock", self)
        self.acDock.setToolTip(text)
        self.acDock.setStatusTip(text)
        self.acDock.triggered.connect(self.onDock)
        self.acDock.setVisible(False)

        text = "Maximize panel"
        self.acMaximize = QAction(icons.maximize, "&Maximize", self)
        self.acMaximize.setToolTip(text)
        self.acMaximize.setStatusTip(text)
        self.acMaximize.triggered.connect(self.onMaximize)

        text = "Minimize panel"
        self.acMinimize = QAction(icons.minimize, "&Minimize", self)
        self.acMinimize.setToolTip(text)
        self.acMinimize.setStatusTip(text)
        self.acMinimize.triggered.connect(self.onMinimize)
        self.acMinimize.setVisible(False)

        self.standard_toolbar = ToolBar("Standard", parent=self)
        self.standard_toolbar.addAction(self.acUndock)
        self.standard_toolbar.addAction(self.acDock)
        self.standard_toolbar.addAction(self.acMaximize)
        self.standard_toolbar.addAction(self.acMinimize)

        # Hidden by default
        self.standard_toolbar.setVisible(False)

    def _build_application_toolbar(self):
        # Creating a separate toolbar that appears in all modes, like
        # Cinema, Concert etc.
        text = "Save LogBook"
        self.acSaveLogBook = QAction(icons.logbook, "&Save LogBook", self)
        self.acSaveLogBook.setToolTip(text)
        self.acSaveLogBook.setStatusTip(text)
        self.acSaveLogBook.triggered.connect(self.handleLogBookPreview)

        self.application_toolbar = ToolBar("Logbook", parent=self)
        self.application_toolbar.add_expander()
        self.application_toolbar.addAction(self.acSaveLogBook)

    def _fill_panel(self):
        # Create the content widget first
        main_content = self.get_content_widget()
        # Then the standard toolbar
        self._build_standard_toolbar()
        self._build_application_toolbar()

        # Build the toolbar container
        toolbar = QWidget(self)
        toolbar_layout = QHBoxLayout(toolbar)
        all_toolbars = self.toolbars() + [self.application_toolbar,
                                          self.standard_toolbar]
        for tb in all_toolbars:
            toolbar_layout.addWidget(tb)
        toolbar_layout.setContentsMargins(0, 0, 0, 0)
        toolbar_layout.setSpacing(0)
        # Make the first toolbars expand to fill all horizontal space
        toolbar_layout.setStretch(toolbar_layout.count() - 2, 1)
        # Default visibility is set to True and updated on container setting
        toolbar.setVisible(True)

        # Setup some visual characteristics of the toolbar container
        stylesheet = 'QWidget#{} {{background-color: rgb(180,180,180); }}'
        name = generateObjectName(toolbar)
        toolbar.setObjectName(name)
        toolbar.setStyleSheet(stylesheet.format(name))
        toolbar.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        # Save a reference so that subclasses can query its properties
        self.toolbar = toolbar

        main_layout = QVBoxLayout(self)
        main_layout.addWidget(self.toolbar)
        main_layout.addWidget(main_content)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)

    def _update_toolbar_buttons(self, action):
        """When a panel is either undocked or maximized, we do not allow any
        action to be performed on it except the one which resets it to its
        default state (either undock or minimize).
        """
        if action is PanelActions.Maximize:
            self.acMaximize.setVisible(False)
            self.acMinimize.setVisible(True)
            self.acDock.setVisible(False)
            self.acUndock.setVisible(False)
        elif action is PanelActions.Minimize:
            self.acMaximize.setVisible(True)
            self.acMinimize.setVisible(False)
            self.acDock.setVisible(False)
            self.acUndock.setVisible(True)
        elif action is PanelActions.Undock:
            self.acDock.setVisible(True)
            self.acUndock.setVisible(False)
            self.acMaximize.setVisible(False)
            self.acMinimize.setVisible(False)
        elif action is PanelActions.Dock:
            self.acDock.setVisible(False)
            self.acUndock.setVisible(True)
            self.acMaximize.setVisible(True)
            self.acMinimize.setVisible(False)
        else:
            raise ValueError('Unrecognized panel action!')



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from collections import namedtuple

from qtpy.QtWidgets import (
    QAction, QDialog, QMessageBox, QStackedLayout, QToolButton, QWidget)
from traits.api import Instance, List

import karabogui.access as krb_access
from karabo.common.api import State
from karabo.common.scenemodel.api import DisplayCommandModel
from karabogui.binding.api import SlotBinding, get_binding_value
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.dialogs.api import FormatLabelDialog
from karabogui.fonts import get_font_size_from_dpi
from karabogui.indicators import LOCKED_COLOR

# An item contains the slotbinding proxy and its connected qt action
Item = namedtuple('Item', ['proxy', 'action'])
DEFAULT_TEXT = 'NO TEXT'

CONFIRM_STYLE = ('QToolButton{{font-size: {}pt; font: bold;}}'
                 'QToolButton:enabled{{color: {};}}')
TOOL_STYLE = 'QToolButton {{ font-size: {}pt; font: {} }}'


@register_binding_controller(ui_name='Command', klassname='DisplayCommand',
                             binding_type=SlotBinding)
class DisplayCommand(BaseBindingController):
    # The scene model class for this controller
    model = Instance(DisplayCommandModel, args=())
    # Internal traits
    _actions = List(Instance(Item))
    _button = Instance(QToolButton)

    def add_proxy(self, proxy):
        action = QAction(DEFAULT_TEXT, self._button)
        action.setEnabled(False)
        self._button.addAction(action)

        item = Item(proxy=proxy, action=action)
        self._actions.append(item)

        if proxy.binding is not None:
            # Only if the binding is already valid
            self._item_creation(item)
        return True

    def remove_proxy(self, proxy):
        """Remove an additional property proxy `proxy` from the controller"""
        item = [item for item in self._actions
                if item.proxy is proxy][0]
        self._button.removeAction(item.action)
        self._actions.remove(item)
        self._set_default_action()
        return True

    def create_widget(self, parent):
        widget = QWidget(parent)

        layout = QStackedLayout(widget)
        self._button = QToolButton(widget)
        layout.addWidget(self._button)

        self.add_proxy(self.proxy)

        confirmation = QAction("Requires Confirmation", widget)
        confirmation.setCheckable(True)
        confirmation.setChecked(self.model.requires_confirmation)
        confirmation.toggled.connect(self._requires_confirmation_slot)

        # Format
        format_action = QAction("Format field...", widget)
        format_action.triggered.connect(self._format_field)

        self._change_button_style(widget)

        widget.addAction(confirmation)
        widget.addAction(format_action)

        return widget

    def destroy_widget(self):
        """Clear the action on destruction of the command widget

        NOTE: This is required since Qt events are handled differently than
        trait events and the actions might have been deleted already when for
        example the binding update is called.
        """
        self._actions.clear()

    def _requires_confirmation_slot(self):
        self.model.requires_confirmation = not self.model.requires_confirmation
        self._change_button_style(self.widget)

    def _change_button_style(self, widget):
        """Receives the widget button which will be modified
        :param widget: QWidget
        """
        font_size = get_font_size_from_dpi(self.model.font_size)
        if self.model.requires_confirmation:
            color = ("rgb({}, {}, {})").format(*LOCKED_COLOR)
            widget.setStyleSheet(CONFIRM_STYLE.format(font_size, color))
        else:
            fw = self.model.font_weight
            widget.setStyleSheet(TOOL_STYLE.format(font_size, fw))

    def binding_update(self, proxy):
        for item in self._actions:
            if item.proxy is proxy:
                initial = item.action.text() == DEFAULT_TEXT
                self._item_creation(item, initial)
                break

    def state_update(self, proxy):
        updated_dev = proxy.root_proxy
        value = get_binding_value(updated_dev.state_binding, '')
        if value == '':
            return
        state = State(value)
        for item in self._actions:
            item_dev = item.proxy.root_proxy
            item_binding = item.proxy.binding
            if item_binding is None:
                # XXX: Deprecated slot on device side, but still included...
                item.action.setEnabled(False)
            elif item_dev is updated_dev:
                is_allowed = item_binding.is_allowed(state)
                is_accessible = (krb_access.GLOBAL_ACCESS_LEVEL >=
                                 item_binding.requiredAccessLevel)
                item.action.setEnabled(is_allowed and is_accessible)

        self._set_default_action()

    def setEnabled(self, enable):
        """Reimplemented to account for access level changes
        """
        for item in self._actions:
            root_proxy = item.proxy.root_proxy
            state_binding = root_proxy.state_binding
            if state_binding is None:
                item.action.setEnabled(False)
                continue
            state = get_binding_value(state_binding, "")
            if state == "":
                continue
            binding = item.proxy.binding
            if binding is None:
                # XXX: Deprecated slot on device side, but still included...
                item.action.setEnabled(False)
                continue
            is_allowed = binding.is_allowed(state)
            is_accessible = (krb_access.GLOBAL_ACCESS_LEVEL >=
                             binding.requiredAccessLevel)
            item.action.setEnabled(is_allowed and is_accessible)

        self._set_default_action()

    def _item_creation(self, item, initial=True):
        """When an item gets its binding or update, finish hooking things up.

        :param initial: true or false for signal connection
        :type initial: bool
        """
        proxy = item.proxy
        display_name = proxy.binding.displayedName or proxy.path
        # if displayed name is not set, use path
        item.action.setText(display_name)
        item.action.setToolTip(proxy.key)
        # only initially we connect signals and slots
        if initial:
            item.action.triggered.connect(
                lambda _: self._dialog_confirmation(item))
            if proxy.root_proxy.state_binding is not None:
                self.state_update(proxy)

    def _dialog_confirmation(self, item_proxy):
        """This slot is necessary when the operator wants a dialog
        confirmation for the user when execute the button action
        :param item_proxy: It's an Item object with proxy and action
        information
        """
        if self.model.requires_confirmation:
            message_box = QMessageBox()
            message_box.setModal(False)
            confirmation = message_box.question(
                self.widget, 'Confirmation',
                'Continue with this operation?',
                QMessageBox.No,
                QMessageBox.Yes)

            if confirmation == QMessageBox.No:
                return
        item_proxy.proxy.execute()

    def _set_default_action(self):
        for a in self._button.actions():
            if a.isEnabled():
                self._button.setDefaultAction(a)
                break
        else:
            self._button.setDefaultAction(self._button.actions()[0])

    # -----------------------------------------------------------------------
    # Formatting methods

    def _format_field(self):
        dialog = FormatLabelDialog(font_size=self.model.font_size,
                                   font_weight=self.model.font_weight,
                                   parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            self.model.trait_set(font_size=dialog.font_size,
                                 font_weight=dialog.font_weight)
            self._change_button_style(self.widget)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtWidgets import QGraphicsTextItem

from karabo.native import Encoding
from karabogui.binding.builder import build_binding
from karabogui.binding.config import apply_configuration
from karabogui.binding.proxy import DeviceProxy, PropertyProxy
from karabogui.controllers.display.tests.image import (
    get_image_hash, get_pipeline_schema)
from karabogui.graph.common.api import Axes
from karabogui.util import process_qt_events

from ..detector_graph import DisplayDetectorGraph, FrameSlider


# class to make available slot for signal test
class DetectorGraphTestSlot:
    def __init__(self):
        self.value = None

    def on_value_changed(self, value):
        self.value = value


def test_detector_graph_signals(gui_app):
    # set up
    axis = DetectorGraphTestSlot()
    cell = DetectorGraphTestSlot()

    frame_slider = FrameSlider()
    frame_slider.set_slider_maximum(10)
    frame_slider.axisChanged.connect(axis.on_value_changed)
    frame_slider.cellChanged.connect(cell.on_value_changed)

    # Test combobox
    assert axis.value is None
    cb_axis = frame_slider.cb_axis
    cb_axis.setCurrentIndex(Axes['Y'].value)
    cb_axis.currentIndexChanged['const QString &'].emit(axis.value)
    assert axis.value == 'Y'

    cb_axis.setCurrentIndex(Axes['X'].value)
    cb_axis.currentIndexChanged['const QString &'].emit(axis.value)
    assert axis.value == 'X'

    # Test slider
    assert cell.value is None
    frame_slider.sl_cell.setSliderPosition(0)
    frame_slider.sl_cell.valueChanged.emit(0)
    assert cell.value == 0
    assert frame_slider.sb_cell.value() == 0

    frame_slider.sl_cell.setSliderPosition(7)
    frame_slider.sl_cell.valueChanged.emit(7)
    assert cell.value == 7
    assert frame_slider.sb_cell.value() == 7

    # Test spinbox
    frame_slider.sb_cell.setValue(0)
    frame_slider.sb_cell.valueChanged.emit(0)
    assert cell.value == 0
    assert frame_slider.sl_cell.value() == 0

    frame_slider.sb_cell.setValue(7)
    frame_slider.sb_cell.valueChanged.emit(7)
    assert cell.value == 7
    assert frame_slider.sl_cell.value() == 7


@pytest.fixture
def detectorGraphTest(gui_app, mocker):
    schema = get_pipeline_schema()
    binding = build_binding(schema)
    root_proxy = DeviceProxy(binding=binding)

    output_proxy = PropertyProxy(root_proxy=root_proxy,
                                 path='output.data')
    img_proxy = PropertyProxy(root_proxy=root_proxy,
                              path='output.data.image')
    controller = DisplayDetectorGraph(proxy=img_proxy)

    mocker.patch.object(QGraphicsTextItem, 'setHtml')
    controller.create(None)
    yield controller, output_proxy
    controller.destroy()


def test_detector_graph_2d_image(detectorGraphTest):
    """When a 2D image is supplied, the detector shouldn't show the
    frameslider"""
    controller, output_proxy = detectorGraphTest
    frame_slider = controller._frame_slider

    # Assert initial state
    assert not frame_slider.isVisible()

    image_hash = get_image_hash(dimZ=1)
    apply_configuration(image_hash, output_proxy.binding)

    process_qt_events()
    process_qt_events()
    # assert frame_slider.isVisible()

    image_hash = get_image_hash()
    apply_configuration(image_hash, output_proxy.binding)

    # assert frame_slider.isVisible()
    assert controller._plot.imageItem.image.shape == (30, 40)
    controller.clear_widget()
    assert controller._plot.imageItem.image.shape == (10, 10)


def test_detector_graph_basics(detectorGraphTest):
    controller, output_proxy = detectorGraphTest
    frame_slider = controller._frame_slider

    # Assert initial state
    assert not frame_slider.isVisible()
    assert controller._axis == Axes.Z
    assert controller._cell == 0

    image_hash = get_image_hash(dimZ=5)
    apply_configuration(image_hash, output_proxy.binding)

    # Viewing through Z axis
    image = controller.widget.plot().imageItem.image
    assert image.shape == (30, 40)
    assert frame_slider.sb_cell.maximum() == 4
    assert frame_slider.sb_cell.value() == 0

    # Let's modify the controller state through the frame slider
    frame_slider.sb_cell.valueChanged.emit(4)
    assert controller._cell == 4

    # Change axis, this should change the viewed image
    frame_slider.axisChanged.emit('Y')
    image = controller.widget.plot().imageItem.image
    assert image.shape == (40, 5)
    assert frame_slider.sb_cell.maximum() == 29
    assert frame_slider.sb_cell.value() == 0


def test_detector_graph_yuv_image(detectorGraphTest):
    controller, output_proxy = detectorGraphTest
    image_hash = get_image_hash(dimZ=3, encoding=Encoding.YUV444)
    apply_configuration(image_hash, output_proxy.binding)

    image_node = controller._image_node
    assert image_node.encoding is Encoding.GRAY
    assert image_node.dim_x == 40
    assert image_node.dim_y == 30
    assert image_node.dim_z == 0



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

from qtpy.QtGui import QValidator

from ..webdialog import WebValidator


def test_valid_url(gui_app):
    validator = WebValidator()
    user_pass_url = "http://username:password@someurl.com"
    ip_address = "192.168.1.1"
    localhost_with_port = "localhost:8080"
    simple_website = "google.com"
    query_url = "ftp://someurl.com/page.html?param1=1&param2=2"
    cinema_url = "karabo://cinema?param1=1&param2=2"
    new_gui_url = "karabo://gui"
    _assert_url(validator, user_pass_url, valid=True)
    _assert_url(validator, ip_address, valid=True)
    _assert_url(validator, localhost_with_port, valid=True)
    _assert_url(validator, simple_website, valid=True)
    _assert_url(validator, query_url, valid=True)
    _assert_url(validator, new_gui_url, valid=True)
    _assert_url(validator, cinema_url, valid=True)
    _assert_url(validator, "invalid url", valid=False)


def _assert_url(validator, url, *, valid):
    result, _, _ = validator.validate(url, pos=0)
    if valid:
        assert result == QValidator.Acceptable
    else:
        assert result != QValidator.Acceptable



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtCore import QSortFilterProxyModel


class TableSortFilterModel(QSortFilterProxyModel):
    def __init__(self, parent=None):
        super().__init__(parent)

    def get_model_data(self, row, column):
        """Relay the request of model data to the source model"""
        index = self.index(row, column)
        source_index = self.mapToSource(index)
        return self.sourceModel().get_model_data(source_index.row(),
                                                 source_index.column())

    def index_ref(self, index):
        source_index = self.mapToSource(index)
        return source_index

    def get_header_key(self, section):
        """Relay the request of header data to the source model"""
        return self.sourceModel().get_header_key(section)

    def add_row(self, row, copy_row=None):
        """Relay request to `sourceModel` to add row"""
        return self.sourceModel().add_row(row, copy_row=copy_row)

    def add_row_below(self, row, copy_row=None):
        """Relay request to `sourceModel` to add row below"""
        return self.sourceModel().add_row_below(row, copy_row=copy_row)

    def moveRow(self, source_parent, source_row, parent, row):
        """Relay request to `sourceModel` for moving a row"""
        return self.sourceModel().moveRow(
            source_parent, source_row, parent, row)

    def remove_row(self, row):
        """Relay request to `sourceModel` for removing row"""
        return self.sourceModel().remove_row(row)

    def move_row_up(self, row):
        """Relay request to `sourceModel` for moving row up"""
        return self.sourceModel().move_row_up(row)

    def move_row_down(self, row):
        """Relay request to `sourceModel` for moving row down"""
        return self.sourceModel().move_row_down(row)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

from qtpy.QtGui import QColor, QFont
from qtpy.QtWidgets import QDialog

from karabo.common.scenemodel.api import StickerModel
from karabogui.dialogs.api import StickerDialog
from karabogui.fonts import get_alias_from_font
from karabogui.testing import click_button


def test_sticker_dialog(gui_app, mocker):
    """Test the sticker dialog"""
    model = StickerModel(
        text="XFEL",
        width=100, height=200)
    dialog = StickerDialog(model)

    assert model != dialog.model
    assert dialog.styleSheet() == ""

    text_widget = dialog.leText
    assert text_widget.width() == model.width
    assert text_widget.height() == model.height
    assert text_widget.toPlainText() == model.text
    assert dialog.pbFont.text() == "Sans Serif, 10pt"

    text = "Source Sans Pro,10,-1,5,50,0,0,0,0,0"
    assert dialog.text_font.toString() == text

    # Only the internal dialog model is modified
    text_widget.setPlainText("XFEL2")
    assert model.text == "XFEL"
    assert dialog.model.text == "XFEL2"

    path = "karabogui.dialogs.sticker_dialog.FontDialog"
    d = mocker.patch(path)
    font = QFont()
    font.setPointSize(20)
    d().qfont = font
    d().exec.return_value = QDialog.Accepted
    click_button(dialog.pbFont)
    assert dialog.model.font == font.toString()

    assert dialog.pbBackground.isEnabled() is True
    assert dialog.model.background == "white"
    click_button(dialog.cbBackground)
    assert dialog.pbBackground.isEnabled() is False
    assert dialog.model.background == "transparent"

    # Enable again to change background
    click_button(dialog.cbBackground)
    path = "karabogui.dialogs.sticker_dialog.QColorDialog"
    d = mocker.patch(path)
    c = QColor(255, 0, 0)
    d.getColor.return_value = c
    click_button(dialog.pbBackground)
    assert dialog.model.background == c.name()
    assert dialog.model.background != "transparent"

    # Change the foreground
    assert dialog.model.foreground == ""
    path = "karabogui.dialogs.sticker_dialog.QColorDialog"
    d = mocker.patch(path)
    c = QColor(0, 255, 0)
    d.getColor.return_value = c
    click_button(dialog.pbTextColor)
    assert dialog.model.foreground == c.name()
    assert dialog.pbFont.text() == "Source Sans Pro, 20pt"


def test_set_text_font_button(gui_app):
    """
    The Font buttons text and font family is updated, on changing the font
    selection. Font size remains same.
    """
    model = StickerModel(
        text="XFEL",
        width=100, height=200)
    dialog = StickerDialog(model)

    font = QFont()
    font.setFamily("Monospaced")
    font.setPointSize(23)
    dialog.text_font = font
    dialog.set_text_font_button()
    button_font = dialog.pbFont.font()

    assert button_font.family() == "Monospaced"
    assert button_font.pointSize() == 10
    text = get_alias_from_font("Monospaced") + ", 23pt"
    assert dialog.pbFont.text() == text

    font.setFamily("Sans Serif")
    font.setPointSize(30)
    dialog.text_font = font
    dialog.set_text_font_button()

    button_font = dialog.pbFont.font()
    assert button_font.family() == "Sans Serif"
    assert button_font.pointSize() == 10
    text = get_alias_from_font("Sans Serif") + ", 30pt"
    assert dialog.pbFont.text() == text

    font.setStrikeOut(True)
    font.setBold(True)
    font.setItalic(True)
    dialog.text_font = font
    dialog.set_text_font_button()

    button_font = dialog.pbFont.font()
    assert button_font.strikeOut()
    assert button_font.bold()
    assert button_font.italic()

    font.setStrikeOut(False)
    font.setBold(False)
    font.setItalic(False)
    dialog.text_font = font
    dialog.set_text_font_button()

    button_font = dialog.pbFont.font()
    assert not button_font.strikeOut()
    assert not button_font.bold()
    assert not button_font.italic()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

from karabo.native import Hash
from karabogui.dialogs.api import ConfigComparisonDialog
from karabogui.testing import click_button


def test_config_dialog_basics(gui_app):
    old = Hash("float", 1.0)
    new = Hash("float", 1.2)
    dialog = ConfigComparisonDialog("Compare", old, new, None)

    assert not dialog._show_comparison
    assert dialog.ui_swap.text() == "Show Comparison"
    assert dialog.ui_config_existing.toPlainText() == "\nfloat\n1.0\n"
    assert dialog.ui_config_new.toPlainText() == "\nfloat\n1.2\n"

    click_button(dialog.ui_swap)
    assert dialog.ui_swap.text() == "Show Configuration"
    assert dialog._show_comparison
    # The value changed as well
    assert dialog.ui_changes.toPlainText() == "\nfloat\n1.2\n"

    # New dialog config
    old = Hash("float", 1.0, "boolProperty", False)
    new = Hash("float", 1.2)
    dialog = ConfigComparisonDialog("Compare", old, new, None)
    assert dialog.ui_changes.toPlainText() == (
        "\nfloat\n1.2\nboolProperty\nValue was removed ...\n")



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import logging

from qtpy.QtCore import QObject, QPoint, Qt, Signal, Slot
from qtpy.QtWidgets import (
    QFrame, QHBoxLayout, QMenu, QPushButton, QTextEdit, QVBoxLayout)

import karabogui.icons as icons

# The main status bar logger entity
_logger = logging.getLogger('KaraboGUI')
_logger.setLevel(logging.INFO)

LOG_HEIGHT = 75
CLEAR_BUTTON_SIZE = 25


class LogMediator(QObject):
    """A mediator class for the `logger` to make sure logging is handled
    gracefully by Qt"""
    signal = Signal(str, logging.LogRecord)


class QtHandler(logging.Handler):
    def __init__(self, slotfunc, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.signaller = LogMediator()
        self.signaller.signal.connect(slotfunc)
        formatter = logging.Formatter(
            fmt='%(asctime)s - %(levelname)s - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S')
        self.setFormatter(formatter)

    def emit(self, record):
        s = self.format(record)
        self.signaller.signal.emit(s, record)


class StatusLogWidget(QFrame):
    """A StatusLog widget holding a QTextEdit widget for log messages

    The log widget can color the log messages according to their severity.
    """
    COLORS = {
        logging.DEBUG: 'Blue',
        logging.INFO: 'Black',
        logging.WARNING: 'DarkOrange',
        logging.ERROR: 'Red',
        logging.CRITICAL: 'Purple',
    }

    def __init__(self, parent=None):
        super().__init__(parent)
        horizontal_layout = QHBoxLayout(self)
        horizontal_layout.setContentsMargins(0, 0, 0, 0)
        self.log_widget = QTextEdit(self)
        self.log_widget.setReadOnly(True)
        self.log_widget.setFixedHeight(LOG_HEIGHT)

        # Configure the log handler
        self.handler = QtHandler(self.update_status)
        logger = get_logger()
        logger.addHandler(self.handler)

        # Transparent background and no focus
        viewport = self.log_widget.viewport()
        viewport.setAutoFillBackground(False)
        self.log_widget.setFocusPolicy(Qt.NoFocus)
        self.log_widget.setContextMenuPolicy(Qt.CustomContextMenu)
        self.log_widget.customContextMenuRequested.connect(
            self._show_context_menu)

        horizontal_layout.addWidget(self.log_widget)
        vertical_layout = QVBoxLayout()
        vertical_layout.setContentsMargins(0, 0, 0, 0)
        vertical_layout.addStretch(1)
        button = QPushButton()
        button.setFixedSize(CLEAR_BUTTON_SIZE, CLEAR_BUTTON_SIZE)
        button.setFocusPolicy(Qt.NoFocus)
        button.setToolTip('Clear log')
        button.setIcon(icons.editClear)
        button.clicked.connect(self.log_widget.clear)
        vertical_layout.addWidget(button)
        horizontal_layout.addLayout(vertical_layout)

    @Slot(str, logging.LogRecord)
    def update_status(self, status, record):
        color = self.COLORS.get(record.levelno, 'black')
        s = f'<font color="{color}">{status}</font>'
        self.log_widget.append(s)
        bar = self.log_widget.verticalScrollBar()
        bar.setValue(bar.maximum())

    @Slot(QPoint)
    def _show_context_menu(self, pos):
        """Show a context menu"""
        menu = QMenu(self.log_widget)
        select_action = menu.addAction('Select All')
        select_action.triggered.connect(self.log_widget.selectAll)
        enable_select = len(self.log_widget.toPlainText()) > 0
        select_action.setEnabled(enable_select)

        copy_action = menu.addAction('Copy Selected')
        copy_action.triggered.connect(self.log_widget.copy)
        enable_copy = not self.log_widget.textCursor().selection().isEmpty()
        copy_action.setEnabled(enable_copy)
        menu.exec(self.log_widget.viewport().mapToGlobal(pos))


def get_logger():
    """Convenience function to get the logger to the gui status"""
    return _logger



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import json

from qtpy.QtGui import QColor

from karabo.native import (
    AccessLevel, AccessMode, Assignment, Configurable, Int8, Node, String,
    Unit)
from karabogui import icons
from karabogui.testing import get_class_property_proxy

from ..utils import (
    dragged_configurator_items, get_child_names, get_icon, get_qcolor_state,
    is_mandatory, threshold_triggered)


class MandatoryNode(Configurable):
    integer = Int8(
        unitSymbol=Unit.METER,
        assignment=Assignment.MANDATORY)


class NormalNode(Configurable):
    integer = Int8(
        unitSymbol=Unit.METER)


class NodeOfNode(Configurable):
    data = Node(MandatoryNode)


class Object(Configurable):
    string = String(
        displayedName='String',
        accessMode=AccessMode.RECONFIGURABLE)

    integer = Int8(
        unitSymbol=Unit.METER)

    nodeMandatory = Node(MandatoryNode)
    nodeNormal = Node(NormalNode)
    nodeOfNode = Node(NodeOfNode)


def test_dragged_configurator_items():
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, 'string')
    mime_data = dragged_configurator_items([proxy])
    items_data = mime_data.data('tree_items').data()
    drop = json.loads(items_data.decode())

    expected = {
        'label': 'String',
        'key': 'string',
        'type': 'Leaf',
        'display_widget_class': 'DisplayLabel',
        'edit_widget_class': 'EditableLineEdit'
    }
    assert drop == [expected]


def test_get_child_names():
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, 'integer')

    property_names = get_child_names(proxy, AccessLevel.EXPERT)
    # Attributes are empty for this proxy
    assert property_names == []

    device_names = get_child_names(proxy.root_proxy, AccessLevel.EXPERT)
    assert device_names == ['string', 'integer', 'nodeMandatory',
                            'nodeNormal', 'nodeOfNode']


def test_get_mandatory():
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, 'integer')
    assert is_mandatory(proxy.binding) is False

    proxy = get_class_property_proxy(schema, 'nodeNormal.integer')
    assert is_mandatory(proxy.binding) is False

    proxy = get_class_property_proxy(schema, 'nodeMandatory.integer')
    assert is_mandatory(proxy.binding) is True

    proxy = get_class_property_proxy(schema, 'nodeOfNode')
    assert is_mandatory(proxy.binding) is True

    proxy = get_class_property_proxy(schema, 'nodeMandatory')
    assert is_mandatory(proxy.binding) is True


def test_get_icon():
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, 'integer')
    assert get_icon(proxy.binding) == icons.int

    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, 'string')
    assert get_icon(proxy.binding) == icons.string


def test_threshold_triggered():
    # check the low limit
    value = 1
    high_limit = None
    low_limit = 2
    assert threshold_triggered(value, low_limit, high_limit) is True
    value = 3
    assert threshold_triggered(value, low_limit, high_limit) is False
    low_limit = None
    assert threshold_triggered(value, low_limit, high_limit) is False

    # check the high limit
    value = 3
    high_limit = 2
    low_limit = 1
    assert threshold_triggered(value, low_limit, high_limit) is True

    high_limit = 3
    assert threshold_triggered(value, low_limit, high_limit) is False
    low_limit = None
    assert threshold_triggered(value, low_limit, high_limit) is False


def test_stateq_color():
    color = get_qcolor_state("UNKNOWN")
    assert isinstance(color, QColor)
    assert color.getRgb() == (255, 200, 150, 128)

    color = get_qcolor_state("banana")
    assert color.getRgb() == (225, 242, 225, 128)

    color = get_qcolor_state("ERROR")
    assert color.getRgb() == (255, 155, 155, 128)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import argparse
import sys
from collections import OrderedDict

from qtpy.QtWidgets import QApplication, QBoxLayout

import karabogui.binding.api as bindings
from karabo.common.scenemodel.api import BoxLayoutModel, LabelModel, SceneModel
from karabogui import icons
from karabogui.controllers.api import (
    get_class_const_trait, get_compatible_controllers, get_scene_model_class,
    populate_controller_registry)
from karabogui.panels.scenepanel import ScenePanel
from karabogui.singletons.api import get_manager, get_network


def _create_scene_model(devs, controllers=None):
    """ Creates a Scene model from registry """

    # these should match the types in the DataGenerator Device
    type_map = OrderedDict()
    type_map['state'] = (bindings.StringBinding, 'State', False)
    type_map['output.schema.node.image'] = (bindings.ImageBinding, '', False)
    type_map['boolProperty'] = (bindings.BoolBinding, '', True)
    type_map['boolPropertyReadOnly'] = (bindings.BoolBinding, '', False)
    type_map['charProperty'] = (bindings.CharBinding, '', True)
    type_map['charPropertyReadOnly'] = (bindings.CharBinding, '', False)
    type_map['int8Property'] = (bindings.Int8Binding, '', True)
    type_map['int8PropertyReadOnly'] = (bindings.Int8Binding, '', False)
    type_map['uint8Property'] = (bindings.Uint8Binding, '', True)
    type_map['uint8PropertyReadOnly'] = (bindings.Uint8Binding, '', False)
    type_map['int16Property'] = (bindings.Int16Binding, '', True)
    type_map['int16PropertyReadOnly'] = (bindings.Int16Binding, '', False)
    type_map['uint16Property'] = (bindings.Uint16Binding, '', True)
    type_map['uint16PropertyReadOnly'] = (bindings.Uint16Binding, '', False)
    type_map['int32Property'] = (bindings.Int32Binding, '', True)
    type_map['int32PropertyReadOnly'] = (bindings.Int32Binding, '', False)
    type_map['uint32Property'] = (bindings.Uint32Binding, '', True)
    type_map['uint32PropertyReadOnly'] = (bindings.Uint32Binding, '', False)
    type_map['int64Property'] = (bindings.Int64Binding, '', True)
    type_map['int64PropertyReadOnly'] = (bindings.Int64Binding, '', False)
    type_map['uint64Property'] = (bindings.Uint64Binding, '', True)
    type_map['uint64PropertyReadOnly'] = (bindings.Uint64Binding, '', False)
    type_map['floatProperty'] = (bindings.FloatBinding, '', True)
    type_map['floatPropertyReadOnly'] = (bindings.FloatBinding, '', False)
    type_map['doubleProperty'] = (bindings.FloatBinding, '', True)
    type_map['doublePropertyReadOnly'] = (bindings.FloatBinding, '', False)
    type_map['stringProperty'] = (bindings.StringBinding, '', True)
    type_map['vectors.boolProperty'] = (bindings.VectorBoolBinding, '', True)
    type_map['vectors.int8Property'] = (bindings.VectorInt8Binding, '', True)
    type_map['vectors.uint8Property'] = (bindings.VectorUint8Binding, '', True)
    type_map['vectors.int16Property'] = (bindings.VectorInt16Binding, '', True)
    type_map['vectors.uint16Property'] = (
        bindings.VectorUint16Binding, '', True)
    type_map['vectors.int32Property'] = (bindings.VectorInt32Binding, '', True)
    type_map['vectors.uint32Property'] = (
        bindings.VectorUint32Binding, '', True)
    type_map['vectors.int64Property'] = (bindings.VectorInt64Binding, '', True)
    type_map['vectors.uint64Property'] = (
        bindings.VectorUint64Binding, '', True)
    type_map['vectors.floatProperty'] = (bindings.VectorFloatBinding, '', True)
    type_map['vectors.doubleProperty'] = (
        bindings.VectorFloatBinding, '', True)
    type_map['vectors.stringProperty'] = (
        bindings.VectorStringBinding, '', True)
    type_map['table'] = (bindings.VectorHashBinding, '', True)

    rows = []
    all_controllers = set()
    for device_id in devs:
        for device_key, binding_data in type_map.items():
            binding, displayType, can_edit = binding_data
            klasses = get_compatible_controllers(binding())
            klasses += get_compatible_controllers(binding(), can_edit=True)
            klasses.sort(key=lambda x: x.__name__)
            all_controllers.update(klass.__name__ for klass in klasses)
            if controllers:
                klasses = [klass for klass in klasses
                           if klass.__name__ in controllers]
            # add the label
            key = '.'.join([device_id, device_key])
            label = LabelModel(text=key)
            row = [label]
            for klass in klasses:
                model_klass = get_scene_model_class(klass)
                class_label = LabelModel(text=klass.__name__)
                traits = {'keys': [key]}
                if get_class_const_trait(klass, '_can_edit'):
                    traits['parent_component'] = 'EditableApplyLaterComponent'
                if 'klass' in model_klass().editable_traits():
                    traits['klass'] = get_class_const_trait(klass,
                                                            '_klassname')
                widget_model = model_klass(**traits)
                model = BoxLayoutModel(direction=QBoxLayout.TopToBottom,
                                       children=[class_label, widget_model])
                row.append(model)
            box = BoxLayoutModel(direction=QBoxLayout.LeftToRight,
                                 children=row)
            rows.append(box)
    vbox = BoxLayoutModel(direction=QBoxLayout.TopToBottom, children=rows)
    scene = SceneModel(children=[vbox])

    print('\n'+'='*60+'\n')
    print('here is the list of all controllers')
    for name in all_controllers:
        print(name)
    print('='*60)
    return scene


def main():
    app = QApplication(sys.argv)

    icons.init()  # Very important!
    populate_controller_registry()

    psa = """
============================================================
This is a testbench gui viewer.

 It will generate a scene containing all possible widgets
 for a PropertyTest device.
 Add a PropertyTest device to the topic you will
 connect with instanceID 'x'
============================================================"""
    ap = argparse.ArgumentParser(
        description=psa, formatter_class=argparse.RawDescriptionHelpFormatter)
    ap.add_argument('controllerName', nargs='*',
                    help='List of Controller Names')
    args = ap.parse_args(sys.argv[1:])
    ap.print_help()
    controllers = args.controllerName
    model = _create_scene_model(['x'], controllers)
    model.simple_name = 'test scene'
    panel = ScenePanel(model, True)
    # XXX: A hack to keep the toolbar visible
    panel.toolbar.setVisible(True)
    panel.show()
    panel.resize(1024, 768)

    # XXX: A hack to connect to the GUI Server for testing
    get_manager()
    get_network().connectToServer()

    app.exec()


if __name__ == '__main__':
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import QDateTime, Qt, Slot
from qtpy.QtWidgets import QDialog

from karabogui import icons
from karabogui.events import (
    KaraboEvent, broadcast_event, register_for_broadcasts,
    unregister_from_broadcasts)
from karabogui.singletons.api import get_network

from .utils import get_dialog_ui

ONE_WEEK = "One Week"
ONE_DAY = "One Day"
ONE_HOUR = "One Hour"
TEN_MINUTES = "Ten Minutes"


class ConfigurationFromPastDialog(QDialog):
    def __init__(self, instance_id, parent=None):
        super().__init__(parent)
        filepath = get_dialog_ui('conftime.ui')
        uic.loadUi(filepath, self)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.instance_id = instance_id
        self.setModal(False)
        self.setWindowFlags(self.windowFlags() | Qt.WindowCloseButtonHint)
        self.ui_timepoint.setDateTime(QDateTime.currentDateTime())
        self.ui_instance_id.setText(instance_id)
        self.ui_request.clicked.connect(self._request_configuration)
        self.ui_show_device.clicked.connect(self._show_device)
        self.event_map = {
            KaraboEvent.NetworkConnectStatus: self._event_network
        }
        self.ui_show_device.setIcon(icons.deviceInstance)
        register_for_broadcasts(self.event_map)

    def _event_network(self, data):
        if not data.get("status"):
            self.close()

    @Slot()
    def _show_device(self):
        broadcast_event(KaraboEvent.ShowDevice, {'deviceId': self.instance_id,
                                                 'showTopology': True})

    @Slot()
    def accept(self):
        """The dialog was accepted and we can request a configuration"""
        self._request_configuration()
        super().accept()

    @Slot()
    def _request_configuration(self):
        # Karabo time points are in UTC
        time_point = self.ui_timepoint.dateTime().toUTC()
        # Explicitly specifiy ISODate!
        time = str(time_point.toString(Qt.ISODate))
        preview = self.ui_preview.isChecked()
        get_network().onGetConfigurationFromPast(self.instance_id, time=time,
                                                 preview=preview)

    def done(self, result):
        """Stop listening for broadcast events"""
        unregister_from_broadcasts(self.event_map)
        super().done(result)

    def _get_time_information(self, selected_time_point):
        current_date_time = QDateTime.currentDateTime()
        if selected_time_point == ONE_WEEK:
            # One week
            time_point = current_date_time.addDays(-7)
        elif selected_time_point == ONE_DAY:
            # One day
            time_point = current_date_time.addDays(-1)
        elif selected_time_point == ONE_HOUR:
            # One hour
            time_point = current_date_time.addSecs(-3600)
        elif selected_time_point == TEN_MINUTES:
            # Ten minutes
            time_point = current_date_time.addSecs(-600)

        return time_point

    def _set_timepoint(self, selected_time_point):
        time = self._get_time_information(selected_time_point)
        self.ui_timepoint.setDateTime(time)

    @Slot()
    def on_ui_one_week_clicked(self):
        self._set_timepoint(ONE_WEEK)

    @Slot()
    def on_ui_one_day_clicked(self):
        self._set_timepoint(ONE_DAY)

    @Slot()
    def on_ui_one_hour_clicked(self):
        self._set_timepoint(ONE_HOUR)

    @Slot()
    def on_ui_ten_minutes_clicked(self):
        self._set_timepoint(TEN_MINUTES)



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on December 7, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QObject

from karabo.common.api import WeakMethodRef


class SystemTopologyListener(QObject):
    """ An object which simply listens to KaraboBroadcastEvents concerning the
    system topology and passes them along via a callback method.
    """
    def __init__(self, notification_cb, parent=None):
        super().__init__(parent)
        # Hold a weak reference to a callback bound method
        self.callback = WeakMethodRef(notification_cb, num_args=2)

    def _event_topology(self, data):
        """ Router for incoming broadcasts
        """
        self.callback(data['devices'], data['servers'])



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

import numpy as np
import pytest

from karabogui.graph.common.api import ImageRegion

from ..analyzer import ProfileAnalyzer

X_LENGTH, Y_LENGTH = (5, 4)


@pytest.fixture()
def analyzer(gui_app):
    profiler = ProfileAnalyzer()
    yield profiler


def test_basics(analyzer):
    assert not analyzer.smooth
    assert analyzer.axis_data is None
    assertArrayIsEmpty(analyzer._x_profile)
    assertArrayIsEmpty(analyzer._y_profile)
    assertArrayIsEmpty(analyzer._x_fit)
    assertArrayIsEmpty(analyzer._y_fit)
    assertArrayIsEmpty(analyzer._fit_params)


def test_analyze(analyzer):
    # Check default, fails without axis data
    analyzer.analyze(_get_region(valid=True))
    assertArrayIsEmpty(analyzer._x_profile)
    assertArrayIsEmpty(analyzer._y_profile)
    analyzer.axis_data = np.arange(X_LENGTH)
    analyzer.analyze(_get_region(valid=True), axis=0)
    assertArrayIsNotEmpty(analyzer._x_profile)
    assertArrayIsNotEmpty(analyzer._y_profile)
    analyzer.analyze(_get_region(valid=False))
    assertArrayIsEmpty(analyzer._x_profile)
    assertArrayIsEmpty(analyzer._y_profile)


def test_fit(analyzer):
    # Check default, fails without axis data and profile data
    analyzer.fit()
    assertArrayIsEmpty(analyzer._x_fit)
    assertArrayIsEmpty(analyzer._y_fit)
    assert analyzer._fit_params is None
    analyzer.axis_data = np.arange(X_LENGTH)
    analyzer.fit()
    assertArrayIsEmpty(analyzer._x_fit)
    assertArrayIsEmpty(analyzer._y_fit)
    assert analyzer._fit_params is None
    analyzer.analyze(_get_region(valid=True), axis=0)
    analyzer.fit()
    assertArrayIsNotEmpty(analyzer._x_fit)
    assertArrayIsNotEmpty(analyzer._y_fit)
    assert analyzer._fit_params is not None
    analyzer.analyze(_get_region(valid=False), axis=0)
    analyzer.fit()
    assertArrayIsEmpty(analyzer._x_fit)
    assertArrayIsEmpty(analyzer._y_fit)
    assert analyzer._fit_params is None


def test_get_stats(analyzer):
    # Check default, fails without axis data and profile data
    assert analyzer.stats == {}
    analyzer.axis_data = np.arange(X_LENGTH)
    assert analyzer.stats == {}
    analyzer.analyze(_get_region(valid=True), axis=0)
    analyzer.fit()
    stats = analyzer.stats
    assert stats != {}
    assert stats.get("fwhm") is not None
    analyzer.analyze(_get_region(valid=False), axis=0)
    analyzer.fit()
    stats = analyzer.stats
    assert stats == {}
    assert stats.get("fwhm") is None


def _get_region(valid=True):
    region = ImageRegion()
    if valid:
        image = np.arange(X_LENGTH * Y_LENGTH).reshape(Y_LENGTH, X_LENGTH)
        region = ImageRegion(region=image, region_type=ImageRegion.Area,
                             x_slice=slice(X_LENGTH), y_slice=slice(Y_LENGTH))
        assert region.valid() is valid

    return region


def assertArrayIsEmpty(array):
    np.testing.assert_array_equal(array, [])


def assertArrayIsNotEmpty(array):
    np.testing.assert_raises(AssertionError, np.testing.assert_array_equal,
                             array, [])



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
from .about import AboutDialog
from .alarm_dialog import AlarmDialog
from .app_configuration_dialog import ApplicationConfigurationDialog
from .client_topology import ClientTopologyDialog
from .compare_device_configurations import DeviceSelectorDialog
from .configuration_comparison import ConfigComparisonDialog
from .configuration_from_past import ConfigurationFromPastDialog
from .configuration_preview import ConfigurationFromPastPreview
from .data_view_dialog import DataViewDialog
from .debug_run_dialog import DebugRunDialog
from .development_topology import DevelopmentTopologyDialog
from .device_capability import DeviceCapabilityDialog
from .font_dialog import FontDialog
from .format_fmt import FormatFmtDialog
from .format_label import FormatLabelDialog
from .guisessioninfo import GuiSessionInfo
from .init_configuration import (
    InitConfigurationDialog, SaveConfigurationDialog)
from .init_configuration_preview import InitConfigurationPreviewDialog
from .listedit import ListEditDialog
from .log_dialog import LogDialog
from .messagebox import KaraboMessageBox
from .pen_dialogs import PenDialog
from .popup_button_dialog import PopupButtonDialog
from .project_topology_dialog import ProjectTopologyDialog
from .proxies_dialog import ProxiesDialog
from .reactive_login_dialog import UserSessionDialog
from .scene_dialogs import ReplaceDialog, ResizeSceneDialog, SceneItemDialog
from .scene_link_dialog import SceneLinkDialog
from .sticker_dialog import GREY, StickerDialog
from .textdialog import TextDialog
from .time_detail_dialog import RequestTimeDialog
from .topology_device_dialog import TopologyDeviceDialog
from .update_dialog import UpdateDialog
from .utils import get_dialog_ui
from .webdialog import WebDialog, WebValidator



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest.mock import Mock

from karabo.common.api import InstanceStatus
from karabo.native import Configurable, Hash, String
from karabogui.binding.api import ProxyStatus
from karabogui.singletons.mediator import Mediator
from karabogui.testing import singletons, system_hash
from karabogui.topology.system_topology import SystemTopology


class FooClass(Configurable):
    val = String()


def test_project_device():
    network = Mock()
    mediator = Mediator()
    topology = SystemTopology()
    topology.initialize(system_hash())
    with singletons(network=network, topology=topology, mediator=mediator):
        device = topology.get_project_device("divvy",
                                             server_id="swerver",
                                             class_id="FooClass")
        offline_proxy = device._offline_proxy
        online_proxy = device._online_proxy

        # mocked network is not providing schema to the online proxy
        online_proxy.status = ProxyStatus.OFFLINE
        config = Hash("val", "foo")
        device.set_project_config_hash(config)
        assert device._offline_config == config

        schema = FooClass.getClassSchema()
        topology.class_schema_updated("swerver", "FooClass", schema)
        # We are lazy and did not request building although we have a
        # class in the topology

        assert len(offline_proxy.binding.value) == 0

        # Now remove the schema for test purposes
        topology._class_schemas.clear()
        # Refresh schema!
        topology.ensure_proxy_class_schema("divvy", "swerver", "FooClass")
        # The request is sent out but no schema is available
        network.onGetClassSchema.assert_called_with("swerver", "FooClass")
        # Schema arrives and we have a proxy!
        topology.class_schema_updated("swerver", "FooClass", schema)
        assert len(offline_proxy.binding.value) == 1
        assert offline_proxy.binding.value.val.value == "foo"

        config = Hash("val", "bar")
        device.set_project_config_hash(config)
        assert offline_proxy.binding.value.val.value == "bar"

        assert repr(offline_proxy) == "<ProjectDeviceProxy deviceId=divvy>"
        device.start_monitoring()
        assert online_proxy._monitor_count == 1
        device.stop_monitoring()
        assert online_proxy._monitor_count == 0

        # Test instanceInfo update
        online_proxy.topology_node.status = InstanceStatus.ERROR
        assert device.instance_status is InstanceStatus.ERROR



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QAction
from traits.api import Bool, Instance, WeakRef

from karabo.common.scenemodel.api import (
    ImageGraphModel, build_graph_config, restore_graph_config)
from karabo.native import Encoding, Timestamp
from karabogui import icons
from karabogui.binding.api import ImageBinding
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.graph.common.api import AuxPlots
from karabogui.graph.image.api import (
    KaraboImageNode, KaraboImagePlot, KaraboImageView, karabo_default_image)


@register_binding_controller(ui_name="Image Graph",
                             klassname="ImageGraph",
                             binding_type=ImageBinding,
                             priority=10,
                             can_show_nothing=False)
class DisplayImageGraph(BaseBindingController):
    # Our Image Graph Model
    model = Instance(ImageGraphModel, args=())
    grayscale = Bool(True)

    _plot = WeakRef(KaraboImagePlot)
    _image_node = Instance(KaraboImageNode, args=())

    _colormap_action = Instance(QAction)

    def create_widget(self, parent):
        widget = KaraboImageView(parent=parent)
        widget.stateChanged.connect(self._change_model)
        widget.toolTipChanged.connect(self.show_timestamp_tooltip)
        widget.add_colorbar()
        widget.add_picker()
        widget.add_roi()
        widget.add_aux(plot=AuxPlots.ProfilePlot, smooth=True)
        widget.add_aux(plot=AuxPlots.Histogram)

        # Finalize
        widget.add_toolbar()

        # Get a reference for our plotting
        self._plot = widget.plot()

        # QActions
        widget.add_axes_labels_dialog()
        widget.add_transforms_dialog()

        # Restore the model information
        widget.restore(build_graph_config(self.model))

        # Undock - Transient model data
        if self.model.undock:
            widget.deactivate_roi()
        else:
            # Offer undock action
            undock_action = QAction(icons.undock, "Undock", widget)
            undock_action.triggered.connect(self._undock_graph)
            self._plot.vb.add_action(undock_action)

        return widget

    def clear_widget(self):
        self._plot.setData(karabo_default_image())

    # -----------------------------------------------------------------------
    # Qt Slots

    def show_timestamp_tooltip(self):
        image_node = self.proxy.value
        if image_node is None:
            return
        timestamp = image_node.pixels.value.data.timestamp
        if timestamp is None:
            # Can happen when we toggle aux plots
            return
        diff = Timestamp().toTimestamp() - timestamp.toTimestamp()
        self.widget.setToolTip("{} --- Last image received {:.3f} s "
                               "ago".format(self.proxy.key, diff))

    def _change_model(self, content):
        self.model.trait_set(**restore_graph_config(content))

    def value_update(self, proxy):
        image_data = proxy.value
        self._image_node.set_value(image_data)

        if not self._image_node.is_valid:
            return

        array = self._image_node.get_data()

        # Enable/disable some widget features depending on the encoding
        self.grayscale = (self._image_node.encoding == Encoding.GRAY
                          and array.ndim == 2)

        self._plot.setData(array)

    def _undock_graph(self):
        """Undock the graph image but don't offer roi option"""
        model = self.model.clone_traits()
        model.undock = True
        broadcast_event(KaraboEvent.ShowUnattachedController,
                        {"model": model})

    def _grayscale_changed(self, grayscale):
        if grayscale:
            self.widget.add_colorbar()
            self.widget.restore({"colormap": self.model.colormap})
            self.widget.enable_aux()
        else:
            self.widget.remove_colorbar()
            self.widget.disable_aux()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np

from karabo.native import Configurable, VectorInt32
from karabogui.binding.api import PropertyProxy
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..vector_xy_graph import DisplayVectorXYGraph


class Object(Configurable):
    index = VectorInt32()
    value = VectorInt32()
    another_value = VectorInt32()


def test_vector_xy_graph(gui_app):
    """Test the setting of both x and y values in the vector xy graph"""
    # setup
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "index")
    device = proxy.root_proxy
    value_proxy = PropertyProxy(root_proxy=device, path="value")

    controller = DisplayVectorXYGraph(proxy=proxy)
    controller.create(None)
    assert controller.widget is not None
    controller.visualize_additional_property(value_proxy)

    # set values
    index = [1, 2, 3, 4, 5]
    value = [42] * 5
    set_proxy_value(proxy, "index", index)
    set_proxy_value(value_proxy, "value", value)

    curve = controller._curves[value_proxy]
    np.testing.assert_almost_equal(list(curve.xData), index)
    np.testing.assert_almost_equal(list(curve.yData), value)

    # set miss match values, x is shorter
    index = [1, 2, 3, 4]
    value = [42] * 5
    set_proxy_value(proxy, "index", index)
    set_proxy_value(value_proxy, "value", value)

    curve = controller._curves[value_proxy]
    np.testing.assert_almost_equal(list(curve.xData), index)
    np.testing.assert_almost_equal(list(curve.yData), value[:len(index)])

    # x grows, y is larger
    index = [1, 2, 3, 4, 5]
    set_proxy_value(proxy, "index", index)
    np.testing.assert_almost_equal(list(curve.yData), value)

    # y gets smaller, x shrinks
    value = [42] * 4
    set_proxy_value(value_proxy, "value", value)
    np.testing.assert_almost_equal(list(curve.yData), value)
    np.testing.assert_almost_equal(list(curve.xData), index[:len(value)])

    # Add another plot with lesser values than main proxy.
    another_value_proxy = PropertyProxy(root_proxy=device,
                                        path="another_value")
    controller.visualize_additional_property(another_value_proxy)
    another_value = [10] * 2
    set_proxy_value(another_value_proxy, "another_value", another_value)

    curve = controller._curves[another_value_proxy]
    np.testing.assert_almost_equal(list(curve.yData), another_value)
    np.testing.assert_almost_equal(
        list(curve.xData), index[:len(another_value)])

    # Make sure the first curve item is not affected.
    curve = controller._curves[value_proxy]
    np.testing.assert_almost_equal(list(curve.yData), value)
    np.testing.assert_almost_equal(list(curve.xData), index[:len(value)])

    # teardown
    controller.destroy()
    assert controller.widget is None



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on March 12, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtCore import Qt, Slot
from qtpy.QtGui import QKeySequence
from qtpy.QtWidgets import (
    QDialog, QHBoxLayout, QInputDialog, QListWidget, QListWidgetItem,
    QPushButton, QVBoxLayout)

from karabogui import icons, messagebox
from karabogui.binding.api import (
    VectorBoolBinding, get_min_max_size, is_vector_floating, is_vector_integer)

USER_ROLE = Qt.UserRole


class ListEditDialog(QDialog):
    def __init__(self, binding, duplicates_ok=True, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Edit list")
        self.binding = binding
        self.duplicates_ok = duplicates_ok

        self._allowed_choices = {}
        if isinstance(self.binding, VectorBoolBinding):
            self._allowed_choices["False"] = False
            self._allowed_choices["True"] = True

        # Check for possible size limitations!
        self.minSize, self.maxSize = get_min_max_size(binding)

        h_layout = QHBoxLayout(self)
        list_widget = QListWidget(self)
        slot = list_widget.currentItemChanged[QListWidgetItem, QListWidgetItem]
        slot.connect(self._update_buttons)
        list_widget.itemDoubleClicked.connect(self.itemDoubleClicked)
        h_layout.addWidget(list_widget)
        self.list_widget = list_widget
        self.list_widget.setDragDropMode(QListWidget.InternalMove)

        # Button Actions
        v_layout = QVBoxLayout()
        button = QPushButton(icons.add, "", self)
        button.clicked.connect(self._add_clicked)
        v_layout.addWidget(button)

        self.edit_button = QPushButton(icons.edit, "", self)
        self.edit_button.clicked.connect(self._edit_clicked)
        v_layout.addWidget(self.edit_button)

        self.remove_button = QPushButton(icons.delete, "", self)
        self.remove_button.clicked.connect(self._remove_clicked)
        v_layout.addWidget(self.remove_button)

        self.move_up_button = QPushButton(icons.arrowFancyUp, "", self)
        self.move_up_button.clicked.connect(self._move_up_clicked)
        v_layout.addWidget(self.move_up_button)

        self.move_down_button = QPushButton(icons.arrowFancyDown, "", self)
        self.move_down_button.clicked.connect(self._move_down_clicked)
        v_layout.addWidget(self.move_down_button)
        v_layout.addStretch(1)

        button = QPushButton("OK", self)
        button.clicked.connect(self.accept)
        v_layout.addWidget(button)

        button = QPushButton("Cancel", self)
        button.clicked.connect(self.reject)
        v_layout.addWidget(button)
        h_layout.addLayout(v_layout)

    # ----------------------------------------------------------------------
    # Public interface

    def set_list(self, values):
        self.list_widget.clear()
        for value in values:
            self._add_item(value)

        self._update_buttons()

    @property
    def values(self):
        return [self.list_widget.item(index).data(USER_ROLE)
                for index in range(self.list_widget.count())]

    @property
    def string_values(self):
        return ",".join(self.list_widget.item(index).text()
                        for index in range(self.list_widget.count()))

    # ----------------------------------------------------------------------

    def keyPressEvent(self, event):
        """Reimplemented method of `QDialog`"""
        if event.matches(QKeySequence.New):
            item = self.list_widget.currentItem()
            if item is None:
                event.accept()
                return
            index = self.list_widget.currentRow()
            value = item.data(USER_ROLE)
            self._insert_item(index + 1, value)
            event.accept()
            return
        elif event.matches(QKeySequence.Delete):
            self._remove_clicked()
            event.accept()
            return
        return super().keyPressEvent(event)

    # ----------------------------------------------------------------------
    # Private interface

    def _insert_item(self, index, value):
        """Insert and item at `index` with `value`"""
        item = QListWidgetItem(str(value))
        item.setData(USER_ROLE, value)
        self.list_widget.insertItem(index, item)
        self.list_widget.setCurrentItem(item)

    def _add_item(self, value):
        """Add an item to the list widget with `value`"""
        item = QListWidgetItem(str(value))
        item.setData(USER_ROLE, value)
        self.list_widget.addItem(item)

    def _retrieve_value(self, mode=None):
        item = self.list_widget.currentItem()
        value = item.data(USER_ROLE) if item is not None else None

        dialog = QInputDialog.getText
        title = "Text"
        if is_vector_floating(self.binding):
            dialog = QInputDialog.getDouble
            title = "Float"
        elif is_vector_integer(self.binding):
            dialog = QInputDialog.getInt
            title = "Integer"

        if value is None:
            value, ok = dialog(self, title, "Add")
        elif dialog == QInputDialog.getText:
            value, ok = dialog(self, title, mode, text=value)
        elif dialog == QInputDialog.getInt:
            value = int(value)
            value, ok = dialog(self, title, mode, value)
        elif dialog == QInputDialog.getDouble:
            value = float(value)
            value, ok = dialog(self, title, mode, value, decimals=3)

        if ok:
            return value

    def _retrieve_choice(self, mode):
        text = ""
        if self.list_widget.currentItem() is not None:
            text = self.list_widget.currentItem().text()

        keys = list(self._allowed_choices.keys())
        index = 0
        if text in keys:
            index = keys.index(text)

        text, ok = QInputDialog.getItem(self, "Bool", mode, keys, index, False)
        if ok:
            return self._allowed_choices[text]

    # ----------------------------------------------------------------------
    # Slots

    @Slot(QListWidgetItem)
    def itemDoubleClicked(self, item):
        """If an item is double clicked on the widget, relay editing"""
        self._edit_clicked()

    @Slot()
    def _add_clicked(self):
        if (self.maxSize is not None
                and self.list_widget.count() == self.maxSize):
            messagebox.show_error("The vector size cannot be greater than {}!"
                                  .format(self.maxSize), parent=self)
            return

        if len(self._allowed_choices) < 1:
            value = self._retrieve_value("Add")
        else:
            value = self._retrieve_choice("Add")

        if (value is None or not self.duplicates_ok and
                self.list_widget.findItems(str(value),
                                           Qt.MatchCaseSensitive)):
            return

        self._add_item(value)
        self.list_widget.setCurrentRow(self.list_widget.count() - 1)
        self._update_buttons()

    @Slot()
    def _edit_clicked(self):
        if len(self._allowed_choices) < 1:
            value = self._retrieve_value("Edit")
        else:
            value = self._retrieve_choice("Edit")

        if (value is None or not self.duplicates_ok and
                self.list_widget.findItems(str(value),
                                           Qt.MatchCaseSensitive)):
            return

        item = self.list_widget.currentItem()
        item.setData(USER_ROLE, value)
        item.setText(str(value))
        self._update_buttons()

    @Slot()
    def _remove_clicked(self):
        if (self.minSize is not None
                and self.list_widget.count() == self.minSize):
            messagebox.show_error("The vector size cannot be smaller than {}!"
                                  .format(self.minSize), parent=self)
            return
        self.list_widget.takeItem(self.list_widget.currentRow())
        self._update_buttons()

    @Slot()
    def _move_up_clicked(self):
        row = self.list_widget.currentRow()
        widget = self.list_widget
        if row > 0:
            item = widget.takeItem(row)
            widget.insertItem(row - 1, item)
            widget.setCurrentRow(row - 1)
            self._update_buttons()

    @Slot()
    def _move_down_clicked(self):
        row = self.list_widget.currentRow()
        widget = self.list_widget
        if row < widget.count() - 1:
            item = widget.takeItem(row)
            widget.insertItem(row + 1, item)
            widget.setCurrentRow(row + 1)
            self._update_buttons()

    @Slot()
    def _update_buttons(self):
        """Enable or disable the `Edit`, `Delete` and `Move` buttons"""
        has_items = self.list_widget.count() > 0
        self.edit_button.setEnabled(has_items)
        self.remove_button.setEnabled(has_items)

        row = self.list_widget.currentRow()
        enable_up = has_items and row > 0
        self.move_up_button.setEnabled(enable_up)
        enable_down = has_items and row < self.list_widget.count() - 1
        self.move_down_button.setEnabled(enable_down)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pathlib import Path

import numpy as np
from pyqtgraph import InfiniteLine
from qtpy import uic
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QDialog
from scipy.optimize import curve_fit

from karabogui import icons
from karabogui.controllers.api import get_array_data
from karabogui.graph.common.api import create_button, make_pen
from karabogui.graph.common.fitting import (
    gaussian_fit, guess_initial_parameters, linear_function_fit, normal_cdf,
    sqsech)
from karabogui.graph.common.formatting import (
    table_body, table_header, table_row)
from karabogui.graph.common.utils import float_to_string
from karabogui.graph.plots.base import KaraboPlotView
from karabogui.graph.plots.utils import generate_baseline

header = """<b><u>{FUNC} Fit Result</b></u>"""
row = table_row(table_body(header="{param}", tabs=("{value}", "{error}")))

FWHM_COEFF = 2 * np.sqrt(2 * np.log(2))
BELL_CURVES = ("Gaussian", "Sech Square")

FIT_FUNCTION_MAP = {
    "Gaussian": {
        "func": gaussian_fit, "param_names": ["Amplitude", "Max Pos", "fwhm"]},
    "Sech Square": {
        "func": sqsech, "param_names": ["Amplitude", "Max Pos", "Center",
                                        "Sigma"]},
    "Step function (CDF)": {
        "func": normal_cdf, "param_names": ["Sigma"]},
    "Linear function": {
        "func": linear_function_fit, "param_names": ["Slope", "Interception"]},
    }


LINE_EDIT_STYLE = """
QLineEdit {color:gray;}
"""


class DataAnalysisDialog(QDialog):
    """
    A dialog to perform curve fitting based on different fit functions.
        - Gaussian function
        - Step function
        - Linear function.
    """

    def __init__(self, proxies, config=None, baseline_proxy=None, parent=None):
        """
        Define the optional baseline_proxy to set the values on X-axis,
        otherwise they are generated based on Y-axis values.

        :param proxies: Proxies to be plotted on Y-axis
        :type proxies: List[PropertyProxy]
        :param config: Graph Configuration
        :type config: Dict
        :param baseline_proxy: Optional Values to be set on X-axis
        :type baseline_proxy: None or PropertyProxy
        :param parent: Parent Widget
        :type parent: QObject
        """
        super().__init__(parent)
        ui_path = Path(__file__).parent / "data_analysis.ui"
        uic.loadUi(ui_path, self)
        self.config = config
        self.x_values = self.y_values = []
        self.baseline_proxy = baseline_proxy

        self._plot_widget = KaraboPlotView()
        self._plot_widget.restore(config)
        self._plot_widget.add_legend()
        self._plot_widget.enable_data_toggle()
        toolbar = self._plot_widget.add_toolbar()
        self._plot_widget.enable_export()
        self.plot_layout.addWidget(self._plot_widget)
        pen = make_pen(color="r")
        self.fit_curve = self._plot_widget.add_curve_item(
            name="fitting", pen=pen)
        self.data_curve = self._plot_widget.add_curve_item(
            name="data")

        self.fit_options_combobox.addItems(list(FIT_FUNCTION_MAP.keys()))
        self.fit_on_update_checkbox.setChecked(False)

        self.auto_update.setChecked(False)
        self.auto_update.toggled.connect(self._enable_auto_update)
        self.fit_button.clicked.connect(self.fit)
        self.update_data_button.clicked.connect(self.update_data)

        self.set_up_property_combobox(proxies)
        self.proxy = proxies[0] if proxies else None
        title = self.proxy.key if proxies else self.baseline_proxy.key
        self.setWindowTitle(f"Data Analysis Dialog|{title}")
        self.update_data()

        self.setup_roi()
        self.set_line_pos()

        self.show_line_roi_button = create_button(
            icon=icons.split, checkable=True,
            tooltip="Fit Data from Selection",
            on_clicked=self._show_vertical_line_roi)
        toolbar.add_button(self.show_line_roi_button)

        self.stackedWidget.setCurrentIndex(0)

        self.left_pos_line_edit.setStyleSheet(LINE_EDIT_STYLE)
        self.right_pos_line_edit.setStyleSheet(LINE_EDIT_STYLE)
        self.on_left_line_moved()
        self.on_right_line_moved()

    def setup_roi(self):
        """Create two vertical lines to define the selection. By default,
        both are hidden"""
        pen = make_pen('o', alpha=255, width=2)
        self.left_line = InfiniteLine(
            angle=90, pen=pen, movable=True, hoverPen=pen)
        self.right_line = InfiniteLine(
            angle=90, pen=pen, movable=True, hoverPen=pen)
        self._plot_widget.plotItem.addItem(
            self.left_line, ignoreBounds=True)
        self._plot_widget.plotItem.addItem(
            self.right_line, ignoreBounds=True)
        self.left_line.sigPositionChangeFinished.connect(
            self.on_left_line_moved)
        self.right_line.sigPositionChangeFinished.connect(
            self.on_right_line_moved)

        self.left_line.setVisible(False)
        self.right_line.setVisible(False)

    def set_line_pos(self):
        """
        Position the lines at first and last points on X-axis. Set their
        bounds by keeping one unit distance between them so that they never
        come at the same position.
        """
        if not len(self.x_values) or not len(self.y_values):
            return
        left = min(self.x_values)
        right = max(self.x_values)
        min_width = self._min_width
        self.left_line.setBounds((left, right - min_width))
        self.right_line.setBounds((left + min_width, right))
        self.left_line.setPos(left)
        self.right_line.setPos(right)
        self.left_pos_line_edit.setText(float_to_string(left))
        self.right_pos_line_edit.setText(float_to_string(right))

    @Slot(bool)
    def _show_vertical_line_roi(self, toggled):
        """ Set visibility of the lines."""
        self.left_line.setVisible(toggled)
        self.right_line.setVisible(toggled)
        self.stackedWidget.setCurrentIndex(int(toggled))

    @Slot()
    def on_left_line_moved(self):
        if not len(self.x_values):
            return
        left_pos = self.left_line.pos().x()
        self.left_pos_line_edit.setText(float_to_string(left_pos))
        bounds = (left_pos + self._min_width, max(self.x_values))
        self.right_line.setBounds(bounds)

    @Slot()
    def on_right_line_moved(self):
        if not len(self.x_values):
            return
        right_pos = self.right_line.pos().x()
        self.right_pos_line_edit.setText(float_to_string(right_pos))
        bounds = (min(self.x_values), right_pos - self._min_width)
        self.left_line.setBounds(bounds)

    @property
    def _min_width(self):
        return (max(self.x_values) - min(self.x_values) + 1) / len(
            self.x_values)

    def set_up_property_combobox(self, proxies: list):
        """
        Add all the available proxies to the combobox as internal
        data and its key as visible text. Also select the item
        corresponding to the proxy which is plotted.

        :param proxies: All available proxies.
        """
        visible = len(proxies) > 1
        self.property_comboBox.setVisible(visible)
        self.property_label.setVisible(visible)

        if not visible:
            return

        for index, proxy in enumerate(proxies):
            self.property_comboBox.addItem(proxy.key)
            self.property_comboBox.setItemData(index, proxy)
        self.property_comboBox.currentIndexChanged.connect(
            self._on_property_changed)

    @Slot(int)
    def _on_property_changed(self):
        """
        Update the plot and dialog's window title with respect to the selected
        proxy.
        """
        proxy = self.property_comboBox.currentData()
        self.proxy = proxy
        self.update_data()
        self.setWindowTitle(f"Data Analysis Dialog|{proxy.key}")
        self.set_line_pos()

    def update_data(self):
        """Fetch the latest data points from the parent plot"""
        if self.proxy is None:
            return
        value, _ = get_array_data(self.proxy)
        if value is None:
            return
        self.y_values = value

        if self.baseline_proxy is None:
            offset = self.config.get("offset", 0.0)
            step = self.config.get("step", 1.0)
            self.x_values = generate_baseline(
                self.y_values, offset=offset, step=step)
        else:
            value, _ = get_array_data(self.baseline_proxy)
            if value is None:
                return
            self.x_values = value

        if len(self.x_values) != len(self.y_values):
            return
        self.data_curve.setData(self.x_values, self.y_values)
        if self.fit_on_update_checkbox.isChecked():
            self.fit()
        else:
            self.fit_curve.setVisible(False)

    @Slot()
    def fit(self):
        """Fit data with the selected function"""
        if not len(self.x_values) or not len(self.y_values):
            return

        x_values = self.x_values
        y_values = self.y_values
        if self.show_line_roi_button.isChecked():
            min_value = self.left_line.getPos()[0]
            max_value = self.right_line.getPos()[0]

            # Slice the data according to position
            size = len(x_values) - 1
            dx = float(x_values[-1] - x_values[0]) / size
            x_min = np.clip(int((min_value - x_values[0]) / dx), 0, size)
            x_max = np.clip(int((max_value - x_values[0]) / dx), 0, size)
            x_values = x_values[x_min:x_max]
            y_values = y_values[x_min:x_max]
        self._fit(x_values, y_values)

    def _fit(self, x_values, y_values):
        xfit = x_values
        fit_option = self.fit_options_combobox.currentText()
        fit_func = FIT_FUNCTION_MAP[fit_option].get("func")
        initial_parameters = None
        if fit_option in BELL_CURVES:
            initial_parameters = guess_initial_parameters(x_values, y_values)
        try:
            params, pcov = curve_fit(f=fit_func, xdata=x_values,
                                     ydata=y_values, p0=initial_parameters)
        except (RuntimeError, TypeError, ValueError):
            xfit = yfit = np.array([])
            params = perr = np.array([])
        else:
            yfit = fit_func(x_values, *params)
            perr = np.sqrt(np.diag(pcov))

        self.fit_curve.setData(xfit, yfit, width=10)

        self._update_fit_result(heading=fit_option, values=params, perr=perr)
        self.fit_curve.setVisible(True)

    def _update_fit_result(self, heading, values, perr):
        """
        Display the best fit params on the  gui.
        """
        text = "<table style='font-size: 11px'>" + header.format(
            FUNC=heading)
        text = text + table_row(table_header(tabs=("Value", "Error")))
        param_names = FIT_FUNCTION_MAP[heading].get("param_names")
        if not values.size:
            values = [np.nan] * len(param_names)
        if not perr.size:
            perr = [np.nan] * len(param_names)

        for param, value, error in zip(param_names, values, perr):
            if param == "fwhm":
                value = FWHM_COEFF * value
            value = float_to_string(value, precision=3)
            error = float_to_string(error, precision=3)
            text = text + row.format(param=param, value=value, error=error)
        text = text + "</table>"
        self.result_label.setText(text)

    def set_axis_labels(self, x_label=None, y_label=None):
        """
        Set the Labels on the X and Y axis items.
        """
        if x_label is not None:
            self._plot_widget.plotItem.setLabel("bottom", x_label)
        if y_label is not None:
            self._plot_widget.plotItem.setLabel("left", y_label)

    @Slot(bool)
    def _enable_auto_update(self, enabled):
        """Update the data in the plot on the parent plot update."""
        self.proxy.on_trait_change(self.update_data,
                                   "binding:config_update", remove=not enabled)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on October 27, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from datetime import datetime

from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QFrame, QHBoxLayout, QMenu, QPushButton, QTextEdit, QVBoxLayout)
from traits.api import Instance

from karabo.common.scenemodel.api import DisplayTextLogModel
from karabo.native import Timestamp
from karabogui import icons
from karabogui.binding.api import StringBinding, get_binding_value
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.generic_scenes import get_property_proxy_model
from karabogui.indicators import ALL_OK_COLOR
from karabogui.util import generateObjectName

BUTTON_SIZE = 32


@register_binding_controller(ui_name='Text Log', klassname='DisplayTextLog',
                             binding_type=StringBinding)
class DisplayTextLog(BaseBindingController):
    # The scene model class for this controller
    model = Instance(DisplayTextLogModel, args=())
    # Internal traits
    log_widget = Instance(QTextEdit)
    _timestamp = Instance(Timestamp, args=())

    def create_widget(self, parent):
        widget = QFrame(parent)
        ver_layout = QVBoxLayout(widget)
        ver_layout.setSpacing(0)

        self.log_widget = QTextEdit(widget)

        # no focus for this widget!
        self.log_widget.setFocusPolicy(Qt.NoFocus)
        self.log_widget.setContextMenuPolicy(Qt.CustomContextMenu)
        self.log_widget.customContextMenuRequested.connect(
            self._show_context_menu)

        ver_layout.addWidget(self.log_widget)
        # start horizontal layout with button and spacer
        hor_layout = QHBoxLayout()

        # strech is required to move the button to the right
        hor_layout.addStretch(1)
        hor_layout.setSpacing(0)

        history_button = QPushButton()
        history_button.setFixedSize(BUTTON_SIZE, BUTTON_SIZE)
        history_button.setFocusPolicy(Qt.NoFocus)
        history_button.setToolTip('Historic Text Log')
        history_button.setIcon(icons.clock)
        history_button.clicked.connect(self.launch_history)
        button = QPushButton()
        button.setFixedSize(BUTTON_SIZE, BUTTON_SIZE)
        button.setFocusPolicy(Qt.NoFocus)
        button.setToolTip('Clear log')
        button.setIcon(icons.editClear)
        button.clicked.connect(self.log_widget.clear)

        hor_layout.addWidget(history_button)
        hor_layout.addWidget(button)

        # spacer item to align button to the right!
        ver_layout.addLayout(hor_layout)

        # nice color background
        objectName = generateObjectName(widget)
        sheet = ('QWidget#{} {{ background-color : rgba{}; }}'
                 ''.format(objectName, ALL_OK_COLOR))
        self.log_widget.setObjectName(objectName)
        self.log_widget.setStyleSheet(sheet)
        return widget

    def value_update(self, proxy):
        # catch both None and empty strings
        value = get_binding_value(proxy, '')
        timestamp = proxy.binding.timestamp
        if value and timestamp != self._timestamp:
            self._write_log(value, timestamp)
            self._timestamp = timestamp

    def _write_log(self, text, timestamp):
        dt = datetime.fromtimestamp(timestamp.toTimestamp())
        stamp = dt.strftime('[%H:%M:%S]')
        item = f'{stamp}: {text}'
        self.log_widget.append(item)

        # update our scroll bar
        bar = self.log_widget.verticalScrollBar()
        bar.setValue(bar.maximum())

    def _show_context_menu(self, pos):
        """Show a context menu"""
        menu = QMenu(self.widget)
        select_action = menu.addAction('Select All')
        select_action.triggered.connect(self.log_widget.selectAll)
        enable_select = len(self.log_widget.toPlainText()) > 0
        select_action.setEnabled(enable_select)

        copy_action = menu.addAction('Copy Selected')
        copy_action.triggered.connect(self.log_widget.copy)
        enable_copy = not self.log_widget.textCursor().selection().isEmpty()
        copy_action.setEnabled(enable_copy)
        menu.exec(self.log_widget.viewport().mapToGlobal(pos))

    def launch_history(self):
        if self.proxy.binding is None:
            return
        model = get_property_proxy_model(self.proxy)
        data = {"model": model}
        broadcast_event(KaraboEvent.ShowUnattachedController, data)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pyqtgraph import LabelItem
from pyqtgraph.graphicsItems.LegendItem import ItemSample
from qtpy.QtCore import Qt
from qtpy.QtGui import QBrush, QColor, QPen

from karabogui.graph.common.api import KaraboLegend, float_to_string


class PickerLegend(KaraboLegend):

    def __init__(self):
        super().__init__()
        self.layout.setContentsMargins(2, 2, 2, 2)
        self.layout.setHorizontalSpacing(15)

        self._color_box = ColorSample(None)
        self.layout.addItem(self._color_box, 0, 0)

        self._label = LabelItem(color='w', size="8pt", justify="left")
        self.layout.addItem(self._label, 0, 1)

        self.layout.setColumnFixedWidth(0, self._color_box.width())
        self.layout.setRowAlignment(0, Qt.AlignVCenter)
        self.layout.setColumnAlignment(1, Qt.AlignLeft)
        self.hide()

    def set_value(self, x, y, value, color):
        # Note: The value of a picked pixel can be either an rgb (array) or
        # a lut value (scalar).
        x = float_to_string(x)
        y = float_to_string(y)
        value = float_to_string(value) if type(value) is float else str(value)
        self._label.setText(f"x: {x}<br>y: {y}<br>value: {value}")
        self._color_box.setBrush(QBrush(color))
        self.updateSize()

    def updateSize(self):
        if self.size is not None:
            return

        self.prepareGeometryChange()
        width = (self._color_box.width() + self._label.width())
        self.setGeometry(0, 0, width, self._label.height())


class ColorSample(ItemSample):
    """The color item in the legend that shows the selected pixel color"""

    def __init__(self, item):
        super().__init__(item)

        self._brush = Qt.NoBrush
        self._pen = QPen(QColor(196, 197, 193, 200), 1)

    def paint(self, painter, *args):
        painter.setPen(self._pen)
        painter.setBrush(self._brush)
        painter.drawRect(0, 0, 12, 45)

    def setBrush(self, brush):
        self._brush = brush
        self.update()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import unittest

import numpy as np

from karabogui.graph.image.roll_image import RollImage


class TestRollImage(unittest.TestCase):
    def test_create(self):
        image = RollImage()
        size = 20
        value = np.array([1] * size)
        image.add(np.arange(size))
        np.testing.assert_array_almost_equal(image.data[0], np.arange(size))
        assert len(image.data) == 1
        image.add(value)
        np.testing.assert_array_almost_equal(image.data[0], value)
        assert len(image.data) == 2

    def test_overflow(self):
        image = RollImage()
        size = 20
        value = np.array([1] * size)
        arange_value = np.arange(size)
        for i in range(image.stack + 1):
            image.add(arange_value)
        np.testing.assert_array_almost_equal(image.data[0], arange_value)
        assert len(image.data) == image.stack
        image.add(value)
        np.testing.assert_array_almost_equal(image.data[0], value)
        assert len(image.data) == image.stack
        np.testing.assert_array_almost_equal(image.data[1], arange_value)

    def test_invalid_values(self):
        image = RollImage()
        size = 20
        value = np.ones(size, dtype=np.float64)
        value[0] = np.inf
        image.add(value)

        # Check if image is finite
        assert np.all(np.isfinite(image.data))
        # Check if inf is turned into zero
        value[0] = 0
        np.testing.assert_equal(image.data[0], value)

        # Add another vector, this time with multiple nans
        value = np.ones(size, dtype=np.float64)
        value[[3, 5]] = np.nan
        image.add(value)

        # Check if image is finite
        assert np.all(np.isfinite(image.data))
        # Check if nans are turned into zero
        value[[3, 5]] = 0
        np.testing.assert_equal(image.data[0], value)

    def test_empty_values(self):
        image = RollImage()
        size = 20

        # Check empty value at first set
        empty = np.array([])
        image.add(empty)
        assert image.data is None

        # Check empty value at nth set
        value = np.ones(size, dtype=np.float64)
        image.add(value)
        image.add(value)
        image.add(value)
        image.add(empty)
        assert image.data is None

    def test_stack_change_reset(self):
        image = RollImage()
        size = 20
        arange_value = np.arange(size)
        for i in range(image.stack + 1):
            image.add(arange_value)
        assert image.data is not None
        # Stack is 100 by default
        image.stack = 100
        assert image.data is not None
        image.stack = 20
        assert image.data is None
        arange_value = np.arange(size)
        for i in range(image.stack + 1):
            image.add(arange_value)
        assert image.data is not None
        image.reset()
        assert image.data is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import warnings
from platform import system

import numpy as np
import pytest

from karabo.common.scenemodel.api import (
    CurveType, VectorGraphModel, extract_graph_curve_option)
from karabo.native import (
    Configurable, Hash, NDArray, UInt32, VectorBool, VectorFloat, VectorInt32)
from karabogui.binding.proxy import PropertyProxy
from karabogui.controllers.display.vector_graph import (
    DisplayNDArrayGraph, DisplayVectorGraph)
from karabogui.graph.plots.utils import create_curve_options, generate_baseline
from karabogui.testing import (
    get_class_property_proxy, set_proxy_hash, set_proxy_value)


class ArrayObject(Configurable):
    prop = NDArray(
        displayedName="NDArray",
        dtype=UInt32,
        shape=(10,))


class VectorObject(Configurable):
    prop = VectorFloat()
    value = VectorInt32(defaultValue=[1, 2, 3])
    bool_value = VectorBool()


def _get_controller_action(controller, text):
    # Get the x-transformation action
    for action in controller.widget.actions():
        if action.text() == text:
            return action


@pytest.fixture
def vector_graph_setup(gui_app):
    # setup
    schema = VectorObject.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    device = proxy.root_proxy
    controller = DisplayVectorGraph(proxy=proxy)
    controller.create(None)
    assert controller.widget is not None
    # yield
    yield controller, proxy, device
    # teardown
    controller.destroy()
    assert controller.widget is None


@pytest.mark.skipif(system() == "Windows",
                    reason="Crosshair does not toggle off with emit==False")
def test_crosshair(vector_graph_setup):
    controller, _, _ = vector_graph_setup
    cross_target = controller.widget._cross_target

    def assert_crosshair_exists(horizontal, vertical):
        assert (cross_target.h_line is not None) is horizontal
        assert (cross_target.v_line is not None) is vertical

    assert_crosshair_exists(False, False)
    cross_target.action_button.clicked.emit(True)
    assert_crosshair_exists(True, True)
    cross_target.action_button.clicked.emit(False)
    assert_crosshair_exists(False, False)


def test_set_value(vector_graph_setup):
    controller, proxy, _ = vector_graph_setup
    assert len(controller._curves) == 1
    curve = controller._curves.get(proxy)
    assert curve is not None
    value = [2, 4, 6]
    set_proxy_value(proxy, "prop", value)
    assert list(curve.yData) == value


def test_set_value_inf(vector_graph_setup):
    controller, proxy, _ = vector_graph_setup
    assert len(controller._curves) == 1
    curve = controller._curves.get(proxy)
    assert curve is not None
    value = [2, 4, 6]
    set_proxy_value(proxy, "prop", value)
    assert list(curve.yData) == value

    value = [2, np.inf, 6]
    set_proxy_value(proxy, "prop", value)
    assert list(curve.yData) == value

    value = [np.inf, np.inf, np.inf]
    set_proxy_value(proxy, "prop", value)

    # None curve in PyQtGraph >= 0.11.1
    assert list(curve.yData) == value

    value = [np.nan, np.nan, np.nan]
    with warnings.catch_warnings(record=True):
        warnings.simplefilter("always")
        set_proxy_value(proxy, "prop", value)
        # Note: Older PyQtGraph version <= 0.11.0 are not catching all
        # NaN values.
        np.testing.assert_almost_equal(list(curve.yData), value)


def test_visualize_prop(vector_graph_setup):
    controller, proxy, device = vector_graph_setup
    value_proxy = PropertyProxy(root_proxy=device, path="value")

    controller.visualize_additional_property(value_proxy)
    assert len(controller._curves) == 2
    assert len(controller.proxies) == 2

    curve = controller._curves.get(value_proxy)
    assert curve is not None
    value = [6, 12, 6]
    set_proxy_value(value_proxy, "value", value)
    assert list(curve.yData) == value

    assert controller.remove_additional_property(value_proxy)
    assert len(controller._curves) == 1
    assert len(controller.proxies) == 1


def test_visualize_curve_options(vector_graph_setup):
    controller, proxy, device = vector_graph_setup
    value_proxy = PropertyProxy(root_proxy=device, path="value")

    controller.visualize_additional_property(value_proxy)
    assert len(controller._curves) == 2
    assert len(controller.proxies) == 2

    curve = controller._curves.get(value_proxy)
    assert curve is not None
    value = [6, 12, 6]
    set_proxy_value(value_proxy, "value", value)
    assert list(curve.yData) == value

    # And now add curve options and alter them

    proxy_curve = controller._curves[proxy]
    assert proxy_curve.name() == "prop"
    assert proxy_curve.opts["pen"].color().name() == "#3399ff"

    options = create_curve_options(
        controller._curves, {}, CurveType.Curve)
    options[proxy_curve]["name"] = "New legend name"
    options[proxy_curve]["pen_color"] = "#ff3333"

    controller.set_curve_options(options)
    assert len(controller.model.curve_options) == 2

    # Verify the changed settings
    assert controller._curves[proxy].name() == "New legend name"
    assert proxy_curve.opts["pen"].color().name() == "#ff3333"

    curve_opt = extract_graph_curve_option(controller.model, value_proxy.key)
    assert len(curve_opt), "Curve options cannot be empty"

    assert controller.remove_additional_property(value_proxy)
    assert len(controller._curves) == 1
    assert len(controller.proxies) == 1

    assert len(controller.model.curve_options) == 1
    curve_opt = extract_graph_curve_option(controller.model, value_proxy.key)
    assert not len(curve_opt), "Curve options must be empty"

    # reset remaining
    controller.reset_curve_options()
    assert controller._curves[proxy].name() == "prop"
    assert proxy_curve.opts["pen"].color().name() == "#3399ff"
    assert len(controller.model.curve_options) == 0


def test_set_bool_value(vector_graph_setup):
    controller, _, device = vector_graph_setup
    bool_value = PropertyProxy(root_proxy=device, path="bool_value")

    controller.visualize_additional_property(bool_value)
    assert len(controller._curves) == 2
    curve = controller._curves.get(bool_value)
    assert curve is not None
    value = [False, False, True]
    set_proxy_value(bool_value, "bool_value", value)
    assert list(curve.yData) == [0, 0, 1]


def test_action_names(vector_graph_setup):
    _, proxy, _ = vector_graph_setup
    controller = DisplayVectorGraph(proxy=proxy, model=VectorGraphModel())
    controller.create(None)
    action = controller.widget.actions()[0]
    assert action.text() == "Grid X"
    action.trigger()
    assert controller.model.x_grid is False
    action.trigger()
    assert controller.model.x_grid is True

    action = controller.widget.actions()[1]
    assert action.text() == "Grid Y"
    action.trigger()
    assert controller.model.y_grid is False
    action.trigger()
    assert controller.model.y_grid is True

    action = controller.widget.actions()[2]
    assert action.text() == "Log X"
    action.trigger()
    assert controller.model.x_log is True
    action.trigger()
    assert controller.model.x_log is False

    action = controller.widget.actions()[3]
    assert action.text() == "Log Y"
    action.trigger()
    assert controller.model.y_log is True
    action.trigger()
    assert controller.model.y_log is False

    action = controller.widget.actions()[4]
    assert action.text() == "Invert X"
    action.trigger()
    assert controller.model.x_invert is True
    action.trigger()
    assert controller.model.x_invert is False

    action = controller.widget.actions()[5]
    assert action.text() == "Invert Y"
    action.trigger()
    assert controller.model.y_invert is True
    action.trigger()
    assert controller.model.y_invert is False

    controller.destroy()


def test_xtransform(vector_graph_setup, mocker):
    controller, proxy, _ = vector_graph_setup
    value = [1, 2, 3]
    set_proxy_value(proxy, "prop", value)

    # Get the action
    action = _get_controller_action(controller, "X-Transformation")
    assert action is not None

    offset = 20
    step = 10
    # Trigger the transformation configuration
    dialog = "karabogui.controllers.basearray.TransformDialog.get"
    content = {"offset": offset, "step": step}
    mocker.patch(dialog, return_value=(content, True))
    action.trigger()

    # Check if model is not
    for proxy, curve in controller._curves.items():
        x_expected = generate_baseline(proxy.value, offset, step)
        np.testing.assert_array_equal(curve.xData, x_expected)


def test_view_settings(vector_graph_setup, mocker):
    controller, _, _ = vector_graph_setup
    # Get the action
    action = _get_controller_action(controller, "View")
    assert action is not None

    # Trigger the transformation configuration
    dialog = "karabogui.graph.plots.dialogs.view.GraphViewDialog.get"
    content = {"title": "TestTitle", "background": "white"}
    mocker.patch(dialog, return_value=(content, True))
    action.trigger()

    config = controller.widget.configuration
    assert config["title"] == "TestTitle"
    assert config["background"] == "white"


def test_arrays_graph_set_value(gui_app):
    # set up
    schema = ArrayObject.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = DisplayNDArrayGraph(proxy=proxy)
    controller.create(None)
    assert controller.widget is not None

    # test body
    curve = controller._curves.get(proxy)
    assert curve is not None
    value = np.array(list(range(10)), dtype="uint32")
    array_hash = Hash("type", 14,
                      "data", value.tobytes())
    h = Hash("prop", array_hash)
    set_proxy_hash(proxy, h)
    np.testing.assert_array_equal(curve.yData, value)

    # teardown
    controller.destroy()
    assert controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.common.project.api import (
    DeviceConfigurationModel, DeviceInstanceModel, DeviceServerModel)
from karabo.native import Hash
from karabogui.project.api import DeviceServerController
from karabogui.testing import singletons, system_hash
from karabogui.topology.api import SystemTopology

from .utils import status_icon_path


def test_device_server_controller(mocker):
    """Test the device server controller"""
    device_icon = mocker.patch(status_icon_path("device"))
    device_icon.return_value = None
    server_icon = mocker.patch(status_icon_path("server"))
    server_icon.return_value = None
    network = mocker.Mock()
    topology = SystemTopology()
    topology.initialize(system_hash())
    with singletons(network=network, topology=topology):
        # 1. Foo device with config
        foo_config = DeviceConfigurationModel(
            class_id="BazClass",
            configuration=Hash("value", "1"))
        foo_config.initialized = True

        foo_model = DeviceInstanceModel(
            class_id="BazClass",
            instance_id="B/B/fooDevice1",
            server_id="testServer",
            configs=[foo_config])
        foo_model.initialized = True
        foo_model.active_config_ref = foo_config.uuid

        # 2. Bar device with config
        bar_config = DeviceConfigurationModel(
            class_id="BazClass",
            configuration=Hash("value", "2"))
        bar_config.initialized = True

        bar_model = DeviceInstanceModel(
            class_id="BazClass",
            server_id="testServer",
            instance_id="A/A/fooDevice2",
            configs=[bar_config])
        bar_model.initialized = True
        bar_model.active_config_ref = bar_config.uuid

        # 3. Server
        server_model = DeviceServerModel(
            server_id="testServer", host="serverFoo",
            devices=[foo_model, bar_model])
        server_model.initialized = True

        # 4. Controllers
        server_controller = DeviceServerController(
            model=server_model)

        assert server_controller.model.devices[0] == foo_model
        assert server_controller.model.devices[1] == bar_model

        server_controller._sort_alphabetically(None)

        # bar starts with A, foo with B
        assert server_controller.model.devices[0] == bar_model
        assert server_controller.model.devices[1] == foo_model

        server_controller._sort_devices_naming(2, None)

        # Sort by member, foo has 1, bar has 2
        assert server_controller.model.devices[0] == foo_model
        assert server_controller.model.devices[1] == bar_model

        server_controller._sort_devices_naming(1, None)

        # Sort by type, foo has B, bar has A
        assert server_controller.model.devices[0] == bar_model
        assert server_controller.model.devices[1] == foo_model

        server_controller._sort_devices_naming(0, None)

        # Sort by domain, but here nothing happens! bar starts with A
        # and foo with B
        assert server_controller.model.devices[0] == bar_model
        assert server_controller.model.devices[1] == foo_model

        # sort by member again foo has 1, bar has 2
        server_controller._sort_devices_naming(2, None)
        assert server_controller.model.devices[0] == foo_model
        assert server_controller.model.devices[1] == bar_model

        # Then sort by domain, changes are happening.
        server_controller._sort_devices_naming(0, None)
        assert server_controller.model.devices[0] == bar_model
        assert server_controller.model.devices[1] == foo_model



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pathlib import Path
from unittest import mock

from qtpy.QtCore import QEvent, QPoint, Qt
from qtpy.QtGui import QMouseEvent
from qtpy.QtWidgets import QDialog

from karabo.common.scenemodel.api import (
    DeviceSceneLinkModel, ImageRendererModel, InstanceStatusModel, SceneModel)
from karabogui import icons
from karabogui.sceneview.api import SceneView
from karabogui.testing import GuiTestCase

from ..drawing import (
    DeviceSceneLinkTool, ImageRendererTool, InstanceStatusTool)


class TestDrawingTools(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.scene_model = SceneModel()
        self.scene_view = SceneView(model=self.scene_model)

    def tearDown(self):
        super().tearDown()
        self.scene_view.destroy()
        self.scene_view = None
        self.scene_model = None

    def test_svg_renderer_tool(self):
        self.assertEqual(len(self.scene_model.children), 0)
        dialog_path = "karabogui.sceneview.tools.drawing.getOpenFileName"
        with mock.patch(dialog_path) as dia:
            path = str(Path(icons.__file__).parent / "about.svg")
            dia.return_value = path

            tool = ImageRendererTool()

            event = QMouseEvent(
                QEvent.MouseMove,
                QPoint(20, 0),
                Qt.LeftButton,
                Qt.LeftButton,
                Qt.NoModifier)

            tool.mouse_up(self.scene_view, event)

            self.assertEqual(len(self.scene_model.children), 1)
            model = self.scene_model.children[0]
            self.assertIsInstance(model, ImageRendererModel)
            self.assertEqual(model.x, 20)
            self.assertEqual(model.y, 0)

    def test_device_scene_link(self):
        self.assertEqual(len(self.scene_model.children), 0)
        path = "karabogui.sceneview.tools.drawing.DeviceCapabilityDialog"
        with mock.patch(path) as dia:
            dia().device_id = "TestDevice"
            dia().capa_name = "scene"
            dia().exec.return_value = QDialog.Accepted

            tool = DeviceSceneLinkTool()

            event = QMouseEvent(
                QEvent.MouseMove,
                QPoint(20, 0),
                Qt.LeftButton,
                Qt.LeftButton,
                Qt.NoModifier)

            tool.mouse_up(self.scene_view, event)

            self.assertEqual(len(self.scene_model.children), 1)
            model = self.scene_model.children[0]
            self.assertIsInstance(model, DeviceSceneLinkModel)
            self.assertEqual(model.x, 20)
            self.assertEqual(model.y, 0)
            self.assertEqual(model.target, "scene")
            self.assertEqual(model.keys, ["TestDevice.availableScenes"])

    def test_instance_status(self):
        self.assertEqual(len(self.scene_model.children), 0)
        path = "karabogui.sceneview.tools.drawing.TopologyDeviceDialog"
        with mock.patch(path) as dia:
            dia().device_id = "TestDevice"
            dia().exec.return_value = QDialog.Accepted

            tool = InstanceStatusTool()

            event = QMouseEvent(
                QEvent.MouseMove,
                QPoint(20, 0),
                Qt.LeftButton,
                Qt.LeftButton,
                Qt.NoModifier)

            tool.mouse_up(self.scene_view, event)

            self.assertEqual(len(self.scene_model.children), 1)
            model = self.scene_model.children[0]
            self.assertIsInstance(model, InstanceStatusModel)
            self.assertEqual(model.x, 20)
            self.assertEqual(model.y, 0)
            self.assertEqual(model.height, 30)
            self.assertEqual(model.width, 30)
            self.assertEqual(model.keys, ["TestDevice.deviceId"])



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on February 25, 2019
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import re

from qtpy.QtGui import QValidator


class HexValidator(QValidator):
    def __init__(self, minInc=None, maxInc=None, parent=None):
        super().__init__(parent)
        self.minInc = minInc
        self.maxInc = maxInc

    def is_hex(self, input):
        return all(i in "0123456789abcdefABCDEF" for i in input)

    def validate(self, input, pos):
        """Reimplemented function of `QValidator`"""
        if input in ("+", "-", ""):
            return self.Intermediate, input, pos

        if not (self.is_hex(input) or
                input[0] in "+-" and self.is_hex(input[1:])):
            return self.Invalid, input, pos

        minInc = self.minInc
        if minInc is not None and minInc >= 0 and input.startswith("-"):
            return self.Invalid, input, pos

        maxInc = self.maxInc
        if maxInc is not None and maxInc < 0 and input.startswith("+"):
            return self.Invalid, input, pos

        if ((minInc is None or minInc <= int(input, base=16)) and
                (maxInc is None or int(input, base=16) <= maxInc)):
            return self.Acceptable, input, pos

        return self.Intermediate, input, pos

    def setBottom(self, minInc):
        self.minInc = minInc

    def setTop(self, maxInc):
        self.maxInc = maxInc


class IntValidator(QValidator):
    def __init__(self, minInc=None, maxInc=None, parent=None):
        super().__init__(parent)
        self.minInc = minInc
        self.maxInc = maxInc

    def validate(self, input, pos):
        """Reimplemented function of `QValidator`"""
        if input in ("+", "-", ""):
            return self.Intermediate, input, pos

        if not (input.isdigit() or input[0] in "+-" and input[1:].isdigit()):
            return self.Invalid, input, pos

        minInc = self.minInc
        if minInc is not None and minInc >= 0 and input.startswith("-"):
            return self.Invalid, input, pos

        maxInc = self.maxInc
        if maxInc is not None and maxInc < 0 and input.startswith("+"):
            return self.Invalid, input, pos

        if ((minInc is None or minInc <= int(input)) and
                (maxInc is None or int(input) <= maxInc)):
            return self.Acceptable, input, pos
        else:
            return self.Intermediate, input, pos

    def setBottom(self, minInc):
        """Reimplemented function of `QValidator`"""
        self.minInc = minInc

    def setTop(self, maxInc):
        """Reimplemented function of `QValidator`"""
        self.maxInc = maxInc


class NumberValidator(QValidator):
    def __init__(self, minInc=None, maxInc=None, decimals=-1, parent=None):
        super().__init__(parent)
        self.minInc = minInc
        self.maxInc = maxInc
        self.decimals = decimals

    def validate(self, input, pos):
        """Reimplemented function of `QValidator`"""
        if input in ("+", "-", ""):
            return self.Intermediate, input, pos

        # we might not have standard notation
        if input[-1] in ("+", "-", "e"):
            return self.Intermediate, input, pos
        # we might not have standard notation
        elif input[-1] in (" "):
            return self.Invalid, input, pos

        # check for floating point precision
        if self.decimals != -1:
            input_split = input.split(".")
            if len(input_split) > 1 and len(input_split[1]) > self.decimals:
                return self.Invalid, input, pos

        # try if we can cast input
        try:
            value = float(input)
        except ValueError:
            return self.Invalid, input, pos
        # then check for limits
        if ((self.minInc is None or self.minInc <= value) and
                (self.maxInc is None or value <= self.maxInc)):
            return self.Acceptable, input, pos

        return self.Intermediate, input, pos

    def setBottom(self, minInc):
        """Reimplemented function of `QValidator`"""
        self.minInc = minInc

    def setTop(self, maxInc):
        """Reimplemented function of `QValidator`"""
        self.maxInc = maxInc


class RegexValidator(QValidator):
    """This is a soft generic regex validator that accepts a regex pattern

    Returns `Intermediate` if the pattern does not match, otherwise
    `Acceptable`!
    """
    pattern = None

    def __init__(self, pattern="", parent=None):
        super().__init__(parent)
        self.pattern = re.compile(pattern)

    def setRegex(self, pattern):
        """Set a new regex pattern"""
        self.pattern = re.compile(pattern)

    def validate(self, input, pos):
        """Reimplemented function of `QValidator`"""
        if not self.pattern.match(input):
            return self.Intermediate, input, pos

        return self.Acceptable, input, pos


class RegexListValidator(QValidator):
    """This is a generic list regex validator that accepts a regex pattern

    Every element is validated against the regex pattern

    :param regex: the regex expression
    :param delimiter: the delimiter (default: ".") which splits the input
    :param minSize: The min size of the list (default: None)
    :param maxSize: The max size of the list (default: None)
    """

    def __init__(self, pattern="", delimiter=",", minSize=None, maxSize=None,
                 parent=None):
        super().__init__(parent)
        self.pattern = re.compile(pattern)
        self.delimiter = delimiter
        self.minSize = minSize
        self.maxSize = maxSize

    def setRegex(self, pattern):
        """Set a new regex pattern"""
        self.pattern = re.compile(pattern)

    def validate(self, input, pos):
        """Reimplemented function of `QValidator`"""
        if not input and (self.minSize is None or self.minSize == 0):
            return self.Acceptable, input, pos
        elif not input and self.minSize is not None and self.minSize > 0:
            return self.Intermediate, input, pos
        elif input.startswith(self.delimiter):
            return self.Intermediate, input, pos

        # check for size first
        values = input.split(self.delimiter)
        if self.minSize is not None and len(values) < self.minSize:
            return self.Intermediate, input, pos
        if self.maxSize is not None and len(values) > self.maxSize:
            return self.Intermediate, input, pos

        if input.endswith(self.delimiter):
            return self.Intermediate, input, pos

        for string in values:
            if not self.pattern.match(string):
                return self.Intermediate, input, pos

        return self.Acceptable, input, pos



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from pyqtgraph import (
    AxisItem, ColorMap, GraphicsWidget, ImageItem, SignalProxy, ViewBox)
from qtpy.QtCore import QPoint, Qt, Signal, Slot
from qtpy.QtGui import QTransform
from qtpy.QtWidgets import QDialog, QGraphicsGridLayout, QMenu

from karabogui.fonts import get_qfont
from karabogui.graph.common.api import COLORMAPS
from karabogui.util import generateObjectName, move_to_cursor

from .dialogs.levels import LevelsDialog
from .utils import ensure_finite_levels, get_level_limits, levels_almost_equal

NUM_SAMPLES = 256


class ColorBarWidget(GraphicsWidget):
    levelsChanged = Signal(object)

    def __init__(self, imageItem, parent=None):
        super().__init__(parent=parent)
        self.setObjectName(generateObjectName(self))
        self.imageItem = imageItem

        self.levels = min_level, max_level = [0, NUM_SAMPLES - 1]
        data = np.linspace(min_level, max_level, NUM_SAMPLES)[None, :]

        self.grid_layout = QGraphicsGridLayout(self)
        self.grid_layout.setSpacing(0)
        self.grid_layout.setContentsMargins(0, 40, 0, 0)

        self.vb = ColorViewBox()
        self.vb.menu = self._create_menu()

        self.barItem = ImageItem(parent=self)
        self.barItem.setImage(data)
        self.vb.addItem(self.barItem)
        self.grid_layout.addItem(self.vb, 0, 0)
        self.vb.setYRange(*self.levels, padding=0)

        font = get_qfont()
        font.setPointSize(8)

        self.axisItem = AxisItem(orientation='right', parent=self)
        self.axisItem.setStyle(tickFont=font)
        self.axisItem.linkToView(self.vb)
        self.grid_layout.addItem(self.axisItem, 0, 1)
        self.setLayout(self.grid_layout)

    # ---------------------------------------------------------------------
    # PyQt slots

    @Slot(object)
    def dynamic_event_throttle(self, event):
        levels = event[0]
        if levels is None:
            bar_level = self.imageItem.image.min(), self.imageItem.image.max()
        else:
            bar_level = levels

        # Set the bar levels different, global levels can be `None` for
        # auto scale
        self.set_levels(bar_level)
        self.levelsChanged.emit(levels)

    @Slot()
    def _show_levels_dialog(self):
        levels = self.imageItem.levels
        if levels is None:
            return  # `None` protection
        image_range = self.imageItem.image.min(), self.imageItem.image.max()
        auto_levels = self.imageItem.auto_levels
        default = None if auto_levels else levels
        limits = get_level_limits(self.imageItem.image)

        dialog = LevelsDialog(levels, image_range, auto_levels,
                              limits=limits, parent=self.parent())
        move_to_cursor(dialog)
        proxy = SignalProxy(dialog.levelsPreview, rateLimit=10,
                            slot=self.dynamic_event_throttle)
        accepted = dialog.exec() == QDialog.Accepted
        proxy.disconnect()
        if accepted:
            levels = dialog.levels
        else:
            levels = default

        bar_levels = image_range if levels is None else levels
        self.set_levels(bar_levels)
        self.levelsChanged.emit(levels)

    # ---------------------------------------------------------------------
    # Qt Events

    def mouseDoubleClickEvent(self, event):
        if event.button() == Qt.LeftButton:
            self._show_levels_dialog()
            event.accept()
            return

        super().mouseDoubleClickEvent(event)

    # ---------------------------------------------------------------------
    # Public methods

    def set_colormap(self, cmap):
        lut = ColorMap(*zip(*COLORMAPS[cmap])).getLookupTable(
            alpha=False)
        self.barItem.setLookupTable(lut)

    def set_margins(self, top=None, bottom=None):
        """Sets top and bottom margins of the colorbar. This depend on the
        axis size of the main plot."""
        if top is None:
            _, top, _, _ = self.grid_layout.getContentsMargins()
        if bottom is None:
            _, _, _, bottom = self.grid_layout.getContentsMargins()

        self.grid_layout.setContentsMargins(0, top, 0, bottom)

    def set_levels(self, image_range):
        """Set the mininum and maximum levels on the color bar"""
        image_min, image_max = ensure_finite_levels(image_range)
        # Check if levels and range are almost equal.
        # Only change y-range values if not.
        if levels_almost_equal(self.levels, [image_min, image_max]):
            return
        # Transform the bar item according to the range
        if image_min == image_max:
            # Protect against same levels and simply lower the minimum
            # for a functional colorbar
            image_min -= 1

        scale = (image_max - image_min) / NUM_SAMPLES
        translate = image_min / scale
        transform = QTransform()
        transform.scale(1, scale)
        transform.translate(0, translate)
        self.barItem.setTransform(transform)
        self.vb.setYRange(image_min, image_max, padding=0)
        self.levels = image_range

    # ---------------------------------------------------------------------
    # Private methods

    def _create_menu(self):
        menu = QMenu()
        menu.addAction("Set levels...", self._show_levels_dialog)
        return menu


class ColorViewBox(ViewBox):

    def __init__(self, parent=None):
        super().__init__(parent=parent, enableMenu=False, enableMouse=False)
        self.setFixedWidth(15)
        self.setLimits(xMin=0, xMax=1)
        self.enableAutoRange(enable=False)

        # Menu placeholder!
        self.menu = None

    # ---------------------------------------------------------------------
    # PyQtGraph methods

    def menuEnabled(self):
        return self.menu is not None

    def raiseContextMenu(self, event):
        if self.menu is None:
            return
        pos = event.screenPos()
        self.menu.popup(QPoint(int(pos.x()), int(pos.y())))



import natsort
from qtpy.QtCore import QModelIndex
from traits.api import Undefined

from karabo.native import Hash
from karabogui.binding.api import NodeBinding


def _friendly_repr(binding, value):
    """Return a user-friendly value"""
    if binding is None or value is Undefined:
        return str(value)

    converters = {"hex": hex, "oct": oct, "bin": bin}
    base = binding.displayType
    if base in converters:
        try:
            return converters[base](value)
        except TypeError:
            # value could be an empty string
            return value
    return value


def create_configurator_info(panel):
    """Create the logbook info for the configuration panel"""

    model = panel.model()

    def _get_row_data(parent_index):
        """Recursively yield the row data for a QModelIndex"""
        for row in range(model.rowCount(parent_index)):
            proxy_index = model.index(row, 1, parent_index)
            proxy = model.index_ref(proxy_index)
            # No pipeline data!
            if proxy is None or proxy.pipeline_parent_path:
                continue
            elif isinstance(proxy.binding, NodeBinding):
                yield from _get_row_data(model.index(row, 0, parent_index))
            else:
                yield proxy.path, _friendly_repr(proxy.binding, proxy.value)

    data = Hash({key: value for key, value
                 in _get_row_data(QModelIndex())})
    return data


def create_scene_info(panel):
    """Create the logbook info for the scene panel"""
    scene_view = panel.scene_view
    proxies = [proxy for proxy in scene_view.cached_proxies()
               if not isinstance(proxy.binding, NodeBinding)]
    data = Hash(natsort.natsorted(
        {proxy.key: _friendly_repr(proxy.binding, proxy.value)
         for proxy in proxies}.items()))

    return data
